"""tests for vak.eval.parametric_umap module"""
import pytest

import vak.config
import vak.common.constants
import vak.common.paths
import vak.eval.frame_classification


def assert_eval_output_matches_expected(model_name, output_dir):
    eval_csv = sorted(output_dir.glob(f"eval_{model_name}*csv"))
    assert len(eval_csv) == 1


@pytest.mark.parametrize(
    "model_name, audio_format, spect_format, annot_format",
    [
        ("ConvEncoderUMAP", "cbin", None, "notmat"),
    ],
)
def test_eval_parametric_umap_model(
        model_name,
        audio_format,
        spect_format,
        annot_format,
        specific_config_toml_path,
        tmp_path,
        trainer_table,
):
    output_dir = tmp_path.joinpath(
        f"test_eval_{audio_format}_{spect_format}_{annot_format}"
    )
    output_dir.mkdir()

    keys_to_change = [
        {"table": "eval", "key": "output_dir", "value": str(output_dir)},
        {"table": "eval", "key": "trainer", "value": trainer_table},
    ]

    toml_path = specific_config_toml_path(
        config_type="eval",
        model=model_name,
        audio_format=audio_format,
        annot_format=annot_format,
        spect_format=spect_format,
        keys_to_change=keys_to_change,
    )
    cfg = vak.config.Config.from_toml_path(toml_path)

    vak.eval.parametric_umap.eval_parametric_umap_model(
        model_config=cfg.eval.model.asdict(),
        dataset_config=cfg.eval.dataset.asdict(),
        trainer_config=cfg.eval.trainer.asdict(),
        checkpoint_path=cfg.eval.checkpoint_path,
        output_dir=cfg.eval.output_dir,
        batch_size=cfg.eval.batch_size,
        num_workers=cfg.eval.num_workers,
    )

    assert_eval_output_matches_expected(cfg.eval.model.name, output_dir)


@pytest.mark.parametrize(
    'path_option_to_change',
    [
        {"table": "eval", "key": "checkpoint_path", "value": '/obviously/doesnt/exist/ckpt.pt'},
    ]
)
def test_eval_frame_classification_model_raises_file_not_found(
    path_option_to_change,
        specific_config_toml_path,
    tmp_path,
    trainer_table
):
    """Test that :func:`vak.eval.parametric_umap.eval_parametric_umap_model`
    raises FileNotFoundError when expected"""
    output_dir = tmp_path.joinpath(
        f"test_eval_cbin_notmat_invalid_dataset_path"
    )
    output_dir.mkdir()

    keys_to_change = [
        {"table": "eval", "key": "output_dir", "value": str(output_dir)},
        {"table": "eval", "key": "trainer", "value": trainer_table},
        path_option_to_change,
    ]

    toml_path = specific_config_toml_path(
        config_type="eval",
        model="ConvEncoderUMAP",
        audio_format="cbin",
        annot_format="notmat",
        spect_format=None,
        keys_to_change=keys_to_change,
    )
    cfg = vak.config.Config.from_toml_path(toml_path)
    with pytest.raises(FileNotFoundError):
        vak.eval.parametric_umap.eval_parametric_umap_model(
            model_config=cfg.eval.model.asdict(),
            dataset_config=cfg.eval.dataset.asdict(),
            trainer_config=cfg.eval.trainer.asdict(),
            checkpoint_path=cfg.eval.checkpoint_path,
            output_dir=cfg.eval.output_dir,
            batch_size=cfg.eval.batch_size,
            num_workers=cfg.eval.num_workers,
        )


@pytest.mark.parametrize(
    'path_option_to_change',
    [
        {"table": "eval", "key": ["dataset", "path"], "value": '/obviously/doesnt/exist/dataset-dir'},
        {"table": "eval", "key": "output_dir", "value": '/obviously/does/not/exist/output'},
    ]
)
def test_eval_frame_classification_model_raises_not_a_directory(
    path_option_to_change,
        specific_config_toml_path,
    trainer_table,
    tmp_path,
):
    """Test that :func:`vak.eval.parametric_umap.eval_parametric_umap_model`
    raises NotADirectoryError when expected"""
    keys_to_change = [
        path_option_to_change,
        {"table": "eval", "key": "trainer", "value": trainer_table},
    ]

    if path_option_to_change["key"] != "output_dir":
        # need to make sure output_dir *does* exist
        # so we don't detect spurious NotADirectoryError and assume test passes
        output_dir = tmp_path.joinpath(
            f"test_eval_raises_not_a_directory"
        )
        output_dir.mkdir()
        keys_to_change.append(
            {"table": "eval", "key": "output_dir", "value": str(output_dir)}
        )

    toml_path = specific_config_toml_path(
        config_type="eval",
        model="ConvEncoderUMAP",
        audio_format="cbin",
        annot_format="notmat",
        spect_format=None,
        keys_to_change=keys_to_change,
    )
    cfg = vak.config.Config.from_toml_path(toml_path)
    with pytest.raises(NotADirectoryError):
        vak.eval.parametric_umap.eval_parametric_umap_model(
            model_config=cfg.eval.model.asdict(),
            dataset_config=cfg.eval.dataset.asdict(),
            trainer_config=cfg.eval.trainer.asdict(),
            checkpoint_path=cfg.eval.checkpoint_path,
            output_dir=cfg.eval.output_dir,
            batch_size=cfg.eval.batch_size,
            num_workers=cfg.eval.num_workers,
        )



from __future__ import annotations

import torch
import torch.nn as nn


class ConvEncoder(nn.Module):
    """Convolutional encoder,
    used by Parametric UMAP model.
    """

    def __init__(
        self,
        input_shape: tuple[int],
        conv1_filters: int = 32,
        conv2_filters: int = 64,
        conv_kernel_size: int = 3,
        conv_stride: int = 2,
        conv_padding: int = 1,
        n_features_linear: int = 256,
        n_components: int = 2,
    ):
        """Initialize a ConvEncoder instance.

        Parameters
        ----------
        input_shape : tuple
            with 3 elements corresponding to dimensions of spectrogram:
            (channels, frequency bins, time bins).
            I.e., we assume input is a spectrogram and treat it like an image,
            typically with one channel; the rows are frequency bins,
            and the columns are time bins.
        conv1_filters : int
            Number of filters in first convolutional layer. Default is 64.
        conv2_filters : int
            Number of filters in second convolutional layer. Default is 128.
        conv_kernel_size : tuple
            Size of kernels, i.e. filters, in convolutional layers. Default is 3.
        conv_padding : int
            Amount of padding for convolutional layers. Default is 1.
        n_components : int
            Number of components of latent space that encoder maps to. Default is 2.
        """
        super().__init__()

        if len(input_shape) != 3:
            raise ValueError(
                "Expected input_shape with length 3, (channels, height, width), "
                f"but input shape was length {len(input_shape)}. "
                f"Input shape was: {input_shape}"
            )

        self.input_shape = input_shape
        self.num_input_channels = input_shape[0]

        self.conv = nn.Sequential(
            nn.Conv2d(
                in_channels=self.num_input_channels,
                out_channels=conv1_filters,
                kernel_size=conv_kernel_size,
                stride=conv_stride,
                padding=conv_padding,
            ),
            nn.MaxPool2d(2, 2),
            nn.Conv2d(
                in_channels=conv1_filters,
                out_channels=conv2_filters,
                kernel_size=conv_kernel_size,
                stride=conv_stride,
                padding=conv_padding,
            ),
            nn.MaxPool2d(2, 2),
            nn.Flatten(),
        )
        mock_input = torch.rand((1, *input_shape))
        mock_conv_out = self.conv(mock_input)
        in_features = mock_conv_out.shape[-1]

        self.encoder = nn.Sequential(
            nn.Linear(in_features, n_features_linear),
            nn.ReLU(),
            nn.Linear(n_features_linear, n_features_linear),
            nn.ReLU(),
            nn.Linear(n_features_linear, n_components),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv(x)
        x = self.encoder(x)
        return x



import pytest

import vak.config


class TestConfig:
    @pytest.mark.parametrize(
            'tables_to_parse',
            [
                None,
                'prep',
                ['prep'],
            ]
    )
    def test_init_with_real_config(
        self, a_generated_config_dict, tables_to_parse
    ):
        """Test that instantiating Config class works as expected"""
        # this is basically the body of the ``Config.from_config_dict`` function.
        config_kwargs = {}

        if tables_to_parse is None:
            for table_name in a_generated_config_dict:
                config_kwargs[table_name] = vak.config.config.TABLE_CLASSES_MAP[table_name].from_config_dict(
                    a_generated_config_dict[table_name]
                )
        else:
            for table_name in a_generated_config_dict:
                if table_name in tables_to_parse:
                    config_kwargs[table_name] = vak.config.config.TABLE_CLASSES_MAP[table_name].from_config_dict(
                        a_generated_config_dict[table_name]
                    )

        config = vak.config.Config(**config_kwargs)

        assert isinstance(config, vak.config.Config)
        # we already test that config loading works for EvalConfig, et al.,
        # so here we just test that the logic of Config works as expected:
        # we should get an attribute for each top-level table that we pass in;
        # if we don't pass one in, then its corresponding attribute should be None
        for attr in ('eval', 'learncurve', 'predict', 'prep', 'train'):
            if tables_to_parse is not None:
                if attr in a_generated_config_dict and attr in tables_to_parse:
                    assert hasattr(config, attr)
                else:
                    assert getattr(config, attr) is None
            else:
                if attr in a_generated_config_dict:
                    assert hasattr(config, attr)

    @pytest.mark.parametrize(
            'tables_to_parse',
            [
                None,
                'prep',
                ['prep'],
            ]
    )
    def test_from_config_dict_with_real_config(
        self, a_generated_config_dict, tables_to_parse
    ):
        """Test :meth:`Config.from_config_dict`"""
        config = vak.config.Config.from_config_dict(
            a_generated_config_dict, tables_to_parse=tables_to_parse
            )

        assert isinstance(config, vak.config.Config)
        # we already test that config loading works for EvalConfig, et al.,
        # so here we just test that the logic of Config works as expected:
        # we should get an attribute for each top-level table that we pass in;
        # if we don't pass one in, then its corresponding attribute should be None
        for attr in ('eval', 'learncurve', 'predict', 'prep', 'train'):
            if tables_to_parse is not None:
                if attr in a_generated_config_dict and attr in tables_to_parse:
                    assert hasattr(config, attr)
                else:
                    assert getattr(config, attr) is None
            else:
                if attr in a_generated_config_dict:
                    assert hasattr(config, attr)

    @pytest.mark.parametrize(
            'tables_to_parse',
            [
                None,
                'prep',
                ['prep'],
            ]
    )
    def test_from_toml_path(self, a_generated_config_path, tables_to_parse):
        config = vak.config.Config.from_toml_path(
            a_generated_config_path, tables_to_parse=tables_to_parse
            )

        assert isinstance(config, vak.config.Config)

        a_generated_config_dict = vak.config.load._load_toml_from_path(a_generated_config_path)
        # we already test that config loading works for EvalConfig, et al.,
        # so here we just test that the logic of Config works as expected:
        # we should get an attribute for each top-level table that we pass in;
        # if we don't pass one in, then its corresponding attribute should be None
        for attr in ('eval', 'learncurve', 'predict', 'prep', 'train'):
            if tables_to_parse is not None:
                if attr in a_generated_config_dict and attr in tables_to_parse:
                    assert hasattr(config, attr)
                else:
                    assert getattr(config, attr) is None
            else:
                if attr in a_generated_config_dict:
                    assert hasattr(config, attr)

    def test_from_toml_path_raises_when_config_doesnt_exist(self, config_that_doesnt_exist):
        with pytest.raises(FileNotFoundError):
            vak.config.Config.from_toml_path(config_that_doesnt_exist)

    def test_invalid_table_raises(self, invalid_table_config_path):
        with pytest.raises(ValueError):
            vak.config.Config.from_toml_path(invalid_table_config_path)

    def test_invalid_key_raises(self, invalid_key_config_path):
        with pytest.raises(ValueError):
            vak.config.Config.from_toml_path(invalid_key_config_path)

    def test_mutiple_top_level_tables_besides_prep_raises(self, invalid_train_and_learncurve_config_toml):
        """Test that a .toml config with two top-level tables besides ``[vak.prep]`` raises a ValueError
        (in this case ``[vak.train]`` and ``[vak.learncurve]``)"""
        with pytest.raises(ValueError):
            vak.config.Config.from_toml_path(invalid_train_and_learncurve_config_toml)



# Do this here to suppress warnings before we import vak
import logging
import shutil
import warnings

from numba.core.errors import NumbaDeprecationWarning
warnings.simplefilter('ignore', category=NumbaDeprecationWarning)

import pandas as pd
import tomlkit

import vak

from . import constants


logger = logging.getLogger(__name__)


def set_up_source_files_and_csv_files_for_frame_classification_models():
    """Set up source files and csv files
    used when testing functionality for frame classification models.

    This function does the following
    - First, get only config files that have the model family set to "frame_classification"
    - Then for all those config files:
      - Generate spectrograms for all the ones that have "spect_output_dir"
      - Then for all the other ones that have "data_dir", set that option in the config file
    - Then for *all* the config files, run `get_or_make_source_files` (again)
      - to get a source files dataframe
      - and save this to csv
      - and then save it again with a ``'split'`` column added
    """
    # first just get configs we're going to prep later
    configs_to_make_spectrograms = [
        config_metadata
        for config_metadata in constants.CONFIG_METADATA
        if config_metadata.model_family == "frame_classification" and config_metadata.spect_output_dir is not None
    ]

    for config_metadata in configs_to_make_spectrograms:
        spect_output_dir = constants.GENERATED_SPECT_OUTPUT_DIR / config_metadata.spect_output_dir
        spect_output_dir.mkdir(parents=True)

        config_path = constants.GENERATED_TEST_CONFIGS_ROOT / config_metadata.filename
        logger.info(
            f"\nRunning :func:`vak.prep.frame_classification.get_or_make_source_files` to generate data for tests, "
            f"using config:\n{config_path.name}"
        )
        cfg = vak.config.Config.from_toml_path(config_path, tables_to_parse='prep')

        source_files_df: pd.DataFrame = vak.prep.frame_classification.get_or_make_source_files(
            data_dir=cfg.prep.data_dir,
            input_type=cfg.prep.input_type,
            audio_format=cfg.prep.audio_format,
            spect_format=cfg.prep.spect_format,
            spect_params=cfg.prep.spect_params,
            spect_output_dir=spect_output_dir,
            annot_format=cfg.prep.annot_format,
            annot_file=cfg.prep.annot_file,
            labelset=cfg.prep.labelset,
            audio_dask_bag_kwargs=cfg.prep.audio_dask_bag_kwargs,
        )

        # We copy annotation files to spect_output_dir
        # so we can "prep" from that directory later.
        # This means we have repeats of some files still, which is annoying;
        # .not.mat files are about ~1.2K though
        for annot_path in source_files_df['annot_path'].values:
            shutil.copy(annot_path, spect_output_dir)

        csv_path = constants.GENERATED_SOURCE_FILES_CSV_DIR / f'{config_metadata.filename}-source-files.csv'
        source_files_df.to_csv(csv_path, index=False)

        config_toml: dict = vak.config.load._load_toml_from_path(config_path)
        purpose = vak.cli.prep.purpose_from_toml(config_toml, config_path)
        dataset_df: pd.DataFrame = vak.prep.frame_classification.assign_samples_to_splits(
            purpose,
            source_files_df,
            dataset_path=spect_output_dir,
            train_dur=cfg.prep.train_dur,
            val_dur=cfg.prep.val_dur,
            test_dur=cfg.prep.test_dur,
            labelset=cfg.prep.labelset,
        )
        source_files_with_split_csv_path = (
                constants.GENERATED_SOURCE_FILES_WITH_SPLITS_CSV_DIR /
                f'{config_metadata.filename}-source-files-with-split.csv'
        )
        dataset_df.to_csv(source_files_with_split_csv_path)

    configs_to_add_data_dir = [
        config_metadata
        for config_metadata in constants.CONFIG_METADATA
        if config_metadata.model_family == "frame_classification" and config_metadata.data_dir is not None
    ]

    for config_metadata in configs_to_add_data_dir:
        config_path = constants.GENERATED_TEST_CONFIGS_ROOT / config_metadata.filename
        logger.info(
            f"\nRunning :func:`vak.prep.frame_classification.get_or_make_source_files` to generate data for tests, "
            f"using config:\n{config_path.name}"
        )

        with config_path.open("r") as fp:
            tomldoc = tomlkit.load(fp)
        data_dir = constants.GENERATED_TEST_DATA_ROOT / config_metadata.data_dir
        tomldoc['vak']['prep']['data_dir'] = str(data_dir)
        with config_path.open("w") as fp:
            tomlkit.dump(tomldoc, fp)

        cfg = vak.config.Config.from_toml_path(config_path, tables_to_parse='prep')

        source_files_df: pd.DataFrame = vak.prep.frame_classification.get_or_make_source_files(
            data_dir=cfg.prep.data_dir,
            input_type=cfg.prep.input_type,
            audio_format=cfg.prep.audio_format,
            spect_format=cfg.prep.spect_format,
            spect_params=cfg.prep.spect_params,
            spect_output_dir=None,
            annot_format=cfg.prep.annot_format,
            annot_file=cfg.prep.annot_file,
            labelset=cfg.prep.labelset,
            audio_dask_bag_kwargs=cfg.prep.audio_dask_bag_kwargs,
        )

        csv_path = constants.GENERATED_SOURCE_FILES_CSV_DIR / f'{config_metadata.filename}-source-files.csv'
        source_files_df.to_csv(csv_path, index=False)

        config_toml: dict = vak.config.load._load_toml_from_path(config_path)
        purpose = vak.cli.prep.purpose_from_toml(config_toml, config_path)
        dataset_df: pd.DataFrame = vak.prep.frame_classification.assign_samples_to_splits(
            purpose,
            source_files_df,
            dataset_path=data_dir,
            train_dur=cfg.prep.train_dur,
            val_dur=cfg.prep.val_dur,
            test_dur=cfg.prep.test_dur,
            labelset=cfg.prep.labelset,
        )
        source_files_with_split_csv_path = (
                constants.GENERATED_SOURCE_FILES_WITH_SPLITS_CSV_DIR /
                f'{config_metadata.filename}-source-files-with-split.csv'
        )
        dataset_df.to_csv(source_files_with_split_csv_path)

    configs_without_spect_output_or_data_dir_to_change = [
        config_metadata
        for config_metadata in constants.CONFIG_METADATA
        if config_metadata.model_family == "frame_classification" and (
                config_metadata.spect_output_dir is None and config_metadata.data_dir is None
        )
    ]
    for config_metadata in configs_without_spect_output_or_data_dir_to_change:
        config_path = constants.GENERATED_TEST_CONFIGS_ROOT / config_metadata.filename
        if not config_path.exists():
            raise FileNotFoundError(f"{config_path} not found")
        logger.info(
            f"\nRunning :func:`vak.prep.frame_classification.get_or_make_source_files` to generate data for tests, "
            f"using config:\n{config_path.name}"
        )
        cfg = vak.config.Config.from_toml_path(config_path, tables_to_parse='prep')
        source_files_df: pd.DataFrame = vak.prep.frame_classification.get_or_make_source_files(
            data_dir=cfg.prep.data_dir,
            input_type=cfg.prep.input_type,
            audio_format=cfg.prep.audio_format,
            spect_format=cfg.prep.spect_format,
            spect_params=cfg.prep.spect_params,
            spect_output_dir=None,
            annot_format=cfg.prep.annot_format,
            annot_file=cfg.prep.annot_file,
            labelset=cfg.prep.labelset,
            audio_dask_bag_kwargs=cfg.prep.audio_dask_bag_kwargs,
        )

        csv_path = constants.GENERATED_SOURCE_FILES_CSV_DIR / f'{config_metadata.filename}-source-files.csv'
        source_files_df.to_csv(csv_path, index=False)

        config_toml: dict = vak.config.load._load_toml_from_path(config_path)
        purpose = vak.cli.prep.purpose_from_toml(config_toml, config_path)
        dataset_df: pd.DataFrame = vak.prep.frame_classification.assign_samples_to_splits(
            purpose,
            source_files_df,
            dataset_path=cfg.prep.data_dir,
            train_dur=cfg.prep.train_dur,
            val_dur=cfg.prep.val_dur,
            test_dur=cfg.prep.test_dur,
            labelset=cfg.prep.labelset,
        )
        source_files_with_split_csv_path = (
                constants.GENERATED_SOURCE_FILES_WITH_SPLITS_CSV_DIR /
                f'{config_metadata.filename}-source-files-with-split.csv'
        )
        dataset_df.to_csv(source_files_with_split_csv_path)



from . import biosoundsegbench
from .biosoundsegbench import BioSoundSegBench, SplitsMetadata
from .get import get

__all__ = [
    "biosoundsegbench",
    "BioSoundSegBench",
    "get",
    "SplitsMetadata",
]

# TODO: make this a proper registry
DATASETS = {"BioSoundSegBench": BioSoundSegBench}



from .crossentropy import CrossEntropyLoss
from .dice import DiceLoss, dice_loss
from .umap import UmapLoss, umap_loss

__all__ = [
    "CrossEntropyLoss",
    "DiceLoss",
    "dice_loss",
    "UmapLoss",
    "umap_loss",
]



import copy

import numpy as np
import pandas as pd
import pytest

import vak.common  # for constants
import vak.common.files.spect
import vak.common.labels


@pytest.mark.parametrize(
    'labelset, map_background',
    [
        (
            set(list("abcde")),
            True
        ),
        (
            set(list("abcde")),
            False
        ),
        (
            {1, 2, 3, 4, 5, 6},
            True,
        ),
        (
            {1, 2, 3, 4, 5, 6},
            False,
        )
    ]
)
def test_to_map(labelset, map_background):
    labelmap = vak.common.labels.to_map(labelset, map_background=map_background)
    assert isinstance(labelmap, dict)
    if map_background:
        # because map_background=True
        assert len(labelmap) == len(labelset) + 1
    else:
        # because map_background=False
        assert len(labelmap) == len(labelset)


@pytest.mark.parametrize(
    'labels_list, expected_labelset',
    [
        (
            [
                [1, 1, 1, 1, 2, 2, 3, 3, 3],
                [1, 1, 1, 2, 2, 3, 3, 3, 3, 3]
            ],
            {1, 2, 3}
        )
    ]
)
def test_to_set(labels_list, expected_labelset):
    labelset = vak.common.labels.to_set(labels_list)
    assert isinstance(labelset, set)
    assert labelset == expected_labelset


@pytest.mark.parametrize(
    'config_type, model_name, audio_format, spect_format, annot_format',
    [
        ('train', 'TweetyNet', 'cbin', None, 'notmat'),
        ('train', 'TweetyNet', None, 'mat', 'yarden'),
    ]
)
def test_from_df(config_type, model_name, audio_format, spect_format, annot_format,
                 specific_dataset_path, specific_dataset_df):
    df = specific_dataset_df(config_type, model_name, annot_format, audio_format, spect_format)
    dataset_path = specific_dataset_path(config_type, model_name, annot_format, audio_format, spect_format)

    out = vak.common.labels.from_df(df, dataset_path)
    assert isinstance(out, list)
    assert all([isinstance(labels, (np.ndarray, pd.arrays.StringArray)) for labels in out])


INTS_LABELMAP = {str(val): val for val in range(1, 20)}
INTS_LABELMAP_WITH_UNLABELED = copy.deepcopy(INTS_LABELMAP)
INTS_LABELMAP_WITH_UNLABELED[vak.common.constants.DEFAULT_BACKGROUND_LABEL] = 0

DEFAULT_SKIP = (vak.common.constants.DEFAULT_BACKGROUND_LABEL,)


@pytest.mark.parametrize(
    'labelmap, skip',
    [
        ({'a': 1, 'b': 2, 'c': 3, 'd': 4, 'e': 5}, None),
        ({vak.common.constants.DEFAULT_BACKGROUND_LABEL: 0, 'a': 1, 'b': 2, 'c': 3, 'd': 4, 'e': 5}, None),
        ({vak.common.constants.DEFAULT_BACKGROUND_LABEL: 0, 'a': 1, 'b': 2, 'c': 3, 'd': 4, 'e': 5}, (vak.common.constants.DEFAULT_BACKGROUND_LABEL,)),
        (INTS_LABELMAP, None),
        (INTS_LABELMAP_WITH_UNLABELED, (vak.common.constants.DEFAULT_BACKGROUND_LABEL,))
    ]
)
def test_multi_char_labels_to_single_char(labelmap, skip):
    if skip:
        out = vak.common.labels.multi_char_labels_to_single_char(labelmap, skip)
    else:
        # test default skip
        out = vak.common.labels.multi_char_labels_to_single_char(labelmap)

    if skip:
        for skiplabel in skip:
            assert skiplabel in out
        assert all(
            [len(label) == 1
             for label in out.keys()
             if label not in skip]
        )
    else:
        assert all([
            len(label) == 1
            for label in out.keys()
            if label not in DEFAULT_SKIP
        ])



"""Constants used by multiple modules.
Defined here to avoid circular imports.
"""

from functools import partial

import crowsetta
import numpy as np
import soundfile
from evfuncs import load_cbin
from scipy.io import loadmat

# ---- audio files ----
AUDIO_FORMAT_FUNC_MAP = {
    "cbin": load_cbin,
    "flac": soundfile.read,
    "ogg": soundfile.read,
    "wav": soundfile.read,
}

VALID_AUDIO_FORMATS = list(AUDIO_FORMAT_FUNC_MAP.keys())

# ---- spectrogram files ----
SPECT_FORMAT_LOAD_FUNCTION_MAP = {
    "mat": partial(loadmat, squeeze_me=True),
    "npz": np.load,
}
VALID_SPECT_FORMATS = list(SPECT_FORMAT_LOAD_FUNCTION_MAP.keys())
DEFAULT_SPECT_FORMAT = "npz"

# ---- valid types of training data, the $x$ that goes into a network
VALID_X_SOURCES = {"audio", "spect"}

# ---- annotation files ----
VALID_ANNOT_FORMATS = crowsetta.formats.as_list()
NO_ANNOTATION_FORMAT = "none"

# format for timestamps
STRFTIME_TIMESTAMP = "%y%m%d_%H%M%S"

# ---- results, from train / learncurve ----
RESULTS_DIR_PREFIX = "results_"

# ---- output (default) file extensions. Using the `pathlib` name "suffix" ----
ANNOT_CSV_SUFFIX = ".annot.csv"
NET_OUTPUT_SUFFIX = ".output.npz"

# ---- the key for loading the spectrogram matrix from an npz file
# TODO: replace this with vocalpy constants when we move to VocalPy
SPECT_KEY = "s"
TIMEBINS_KEY = "t"

# TODO: replace this with vocalpy extension when we move to VocalPy
# ---- the extension used to save spectrograms in npz array files
# used by :func:`vak.prep.spectrogram_dataset.audio_helper.make
SPECT_NPZ_EXTENSION = ".spect.npz"
SPECT_FORMAT_EXT_MAP = {
    "npz": SPECT_NPZ_EXTENSION,
    "mat": ".mat",
}

VALID_SPLITS = ("predict", "test", "train", "val")

DEFAULT_BACKGROUND_LABEL = "background"



"""Default transforms for frame classification models.

These are "item" transforms because they apply transforms to input parameters
and then return them in an "item" (dictionary)
that is turn returned by the __getitem__ method of a vak.InferDatapipe.
Having the transform return a dictionary makes it possible to avoid
coupling the InferDatapipe __getitem__ implementation to the transforms
needed for specific neural network models, e.g., whether the returned
output includes a mask to crop off padding that was added.
"""

from __future__ import annotations

import torch
import torchvision.transforms

from .. import transforms as vak_transforms
from ..transforms import FramesStandardizer


class TrainItemTransform:
    """Default transform used when training frame classification models"""

    def __init__(
        self,
        frames_standardizer: FramesStandardizer | None = None,
    ):
        if frames_standardizer is not None:
            if isinstance(
                frames_standardizer, vak_transforms.FramesStandardizer
            ):
                frames_transform = [frames_standardizer]
            else:
                raise TypeError(
                    f"invalid type for frames_standardizer: {type(frames_standardizer)}. "
                    "Should be an instance of vak.transforms.StandardizeSpect"
                )
        else:
            frames_transform = []

        frames_transform.extend(
            [
                vak_transforms.ToFloatTensor(),
                vak_transforms.AddChannel(),
            ]
        )
        self.frames_transform = torchvision.transforms.Compose(
            frames_transform
        )
        self.frame_labels_transform = vak_transforms.ToLongTensor()

    def __call__(self, frames, frame_labels, spect_path=None):
        frames = self.frames_transform(frames)
        frame_labels = self.frame_labels_transform(frame_labels)
        item = {
            "frames": frames,
            "multi_frame_labels": frame_labels,
        }

        if spect_path is not None:
            item["spect_path"] = spect_path

        return item


class InferItemTransform:
    """Default transform used when running inference on frame classification models,
    for evaluation or to generate new predictions.

    Returned item includes frames reshaped into a stack of windows,
    with padded added to make reshaping possible.
    Any `frame_labels` are not padded and reshaped,
    but are converted to :class:`torch.LongTensor`.
    If return_padding_mask is True, item includes 'padding_mask' that
    can be used to crop off any predictions made on the padding.

    Attributes
    ----------
    frames_standardizer : vak.transforms.FramesStandardizer
        instance that has already been fit to dataset, using fit_df method.
        Default is None, in which case no standardization transform is applied.
    window_size : int
        width of window in number of elements. Argument to PadToWindow transform.
    frames_padval : float
        Value to pad frames with. Added to end of array, the "right side".
        Argument to PadToWindow transform. Default is 0.0.
    frame_labels_padval : int
        Value to pad frame labels vector with. Added to the end of the array.
        Argument to PadToWindow transform. Default is -1.
        Used with ``ignore_index`` argument of :mod:`torch.nn.CrossEntropyLoss`.
    return_padding_mask : bool
        if True, the dictionary returned by ItemTransform classes will include
        a boolean vector to use for cropping back down to size before padding.
        padding_mask has size equal to width of padded array, i.e. original size
        plus padding at the end, and has values of 1 where
        columns in padded are from the original array,
        and values of 0 where columns were added for padding.
    """

    def __init__(
        self,
        window_size,
        frames_standardizer=None,
        frames_padval=0.0,
        frame_labels_padval=-1,
        return_padding_mask=True,
        channel_dim=1,
    ):
        self.window_size = window_size
        self.frames_padval = frames_padval
        self.frame_labels_padval = frame_labels_padval
        self.return_padding_mask = return_padding_mask
        self.channel_dim = channel_dim

        if frames_standardizer is not None:
            if not isinstance(
                frames_standardizer, vak_transforms.FramesStandardizer
            ):
                raise TypeError(
                    f"Invalid type for frames_standardizer: {type(frames_standardizer)}. "
                    "Should be an instance of vak.transforms.FramesStandardizer"
                )
        self.frames_standardizer = frames_standardizer

        self.pad_to_window = vak_transforms.PadToWindow(
            window_size, frames_padval, return_padding_mask=return_padding_mask
        )

        self.frames_transform_after_pad = torchvision.transforms.Compose(
            [
                vak_transforms.ViewAsWindowBatch(window_size),
                vak_transforms.ToFloatTensor(),
                # below, add channel at first dimension because windows become batch
                vak_transforms.AddChannel(channel_dim=channel_dim),
            ]
        )

        self.frame_labels_padval = frame_labels_padval
        self.frame_labels_transform = vak_transforms.ToLongTensor()

    def __call__(
        self,
        frames: torch.Tensor,
        frame_labels: torch.Tensor | None = None,
        frames_path=None,
    ) -> dict:
        if self.frames_standardizer:
            frames = self.frames_standardizer(frames)

        if self.pad_to_window.return_padding_mask:
            frames, padding_mask = self.pad_to_window(frames)
        else:
            frames = self.pad_to_window(frames)
            padding_mask = None
        frames = self.frames_transform_after_pad(frames)

        item = {
            "frames": frames,
        }

        if frame_labels is not None:
            frame_labels = self.frame_labels_transform(frame_labels)
            item["multi_frame_labels"] = frame_labels

        if padding_mask is not None:
            item["padding_mask"] = padding_mask

        if frames_path is not None:
            # make sure frames_path is a str, not a pathlib.Path
            item["frames_path"] = str(frames_path)

        return item



from unittest import mock

import pandas as pd
import pytest

import vak

FAKE_SOURCE_FILES_DF = pd.DataFrame.from_records(
    [
        {'audio_path': 'bird0-2023.10.12.cbin',
         'spect_path': 'bird0-2023.10.12.cbin.spect.npz',
         'annot_path': 'bird0-2023.10.12.cbin.not.mat'}
    ]
)


@pytest.mark.parametrize(
    'config_type, model_name, audio_format, spect_format, annot_format, input_type',
    [
        ('train', 'TweetyNet', 'cbin', None, 'notmat', 'spect'),
        ('predict', 'TweetyNet', 'cbin', None, 'notmat', 'spect'),
        ('eval', 'TweetyNet', 'cbin', None, 'notmat', 'spect'),
        ('train', 'TweetyNet', None, 'mat', 'yarden', 'spect'),
        ('learncurve', 'TweetyNet', 'cbin', None, 'notmat', 'spect'),
    ]
)
def test_get_or_make_source_files(
        config_type, model_name, audio_format, spect_format, annot_format,
        input_type, tmp_path, specific_config_toml_path
):
    """Test that this `vak.prep.frame_classification.get_or_make_source_files` dispatches correctly.

    Other unit tests already test the functions that this function calls.
    """
    toml_path = specific_config_toml_path(
        config_type,
        model_name,
        annot_format,
        audio_format,
        spect_format,
    )

    cfg = vak.config.Config.from_toml_path(toml_path)

    # ---- set up ----
    tmp_dataset_path = tmp_path / 'dataset_dir'
    tmp_dataset_path.mkdir()

    if cfg.prep.input_type == 'audio':
        with mock.patch('vak.prep.frame_classification.source_files.prep_audio_dataset', autospec=True) as mock_prep_audio_dataset:
            mock_prep_audio_dataset.return_value = FAKE_SOURCE_FILES_DF

            out: pd.DataFrame = vak.prep.frame_classification.get_or_make_source_files(
                cfg.prep.data_dir,
                cfg.prep.input_type,
                cfg.prep.audio_format,
                cfg.prep.spect_format,
                cfg.prep.spect_params,
                tmp_dataset_path,
                cfg.prep.annot_format,
                cfg.prep.annot_file,
                cfg.prep.labelset,
                cfg.prep.audio_dask_bag_kwargs,
            )

            assert mock_prep_audio_dataset.called
            assert isinstance(out, pd.DataFrame)

    elif cfg.prep.input_type == 'spect':
        with mock.patch(
                'vak.prep.frame_classification.source_files.prep_spectrogram_dataset', autospec=True
        ) as mock_prep_spect_dataset:
            mock_prep_spect_dataset.return_value = FAKE_SOURCE_FILES_DF

            out: pd.DataFrame = vak.prep.frame_classification.get_or_make_source_files(
                cfg.prep.data_dir,
                cfg.prep.input_type,
                cfg.prep.audio_format,
                cfg.prep.spect_format,
                cfg.prep.spect_params,
                tmp_dataset_path,
                cfg.prep.annot_format,
                cfg.prep.annot_file,
                cfg.prep.labelset,
                cfg.prep.audio_dask_bag_kwargs,
            )
            assert mock_prep_spect_dataset.called
            assert isinstance(out, pd.DataFrame)



"""Class that represents ``[vak.predict]`` table of configuration file."""

from __future__ import annotations

import os
from pathlib import Path

from attr import converters, validators
from attr.validators import instance_of
from attrs import define, field

from ..common.converters import expanded_user_path
from .dataset import DatasetConfig
from .model import ModelConfig
from .trainer import TrainerConfig

REQUIRED_KEYS = (
    "checkpoint_path",
    "dataset",
    "model",
    "trainer",
)


@define
class PredictConfig:
    """Class that represents ``[vak.predict]`` table of configuration file.

     Attributes
     ----------
    checkpoint_path : str
        path to directory with checkpoint files saved by Torch, to reload model
    labelmap_path : str
        path to 'labelmap.json' file.
    model : vak.config.ModelConfig
        The model to use: its name,
        and the parameters to configure it.
        Must be an instance of :class:`vak.config.ModelConfig`
    batch_size : int
        number of samples per batch presented to models during training.
    dataset : vak.config.DatasetConfig
        The dataset to use: the path to it,
        and optionally a path to a file representing splits,
        and the name, if it is a built-in dataset.
        Must be an instance of :class:`vak.config.DatasetConfig`.
    trainer : vak.config.TrainerConfig
        Configuration for :class:`lightning.pytorch.Trainer`.
        Must be an instance of :class:`vak.config.TrainerConfig`.
    num_workers : int
        Number of processes to use for parallel loading of data.
        Argument to torch.DataLoader. Default is 2.
    frames_standardizer_path : str
        path to a saved :class:`vak.transforms.FramesStandardizer` object used to standardize (normalize) frames.
        If spectrograms were normalized and this is not provided, will give
        incorrect results.
    annot_csv_filename : str
        name of .csv file containing predicted annotations.
        Default is None, in which case the name of the dataset .csv
        is used, with '.annot.csv' appended to it.
    output_dir : str
        path to location where .csv containing predicted annotation
        should be saved. Defaults to current working directory.
    min_segment_dur : float
        minimum duration of segment, in seconds. If specified, then
        any segment with a duration less than min_segment_dur is
        removed from lbl_tb. Default is None, in which case no
        segments are removed.
    majority_vote : bool
        if True, transform segments containing multiple labels
        into segments with a single label by taking a "majority vote",
        i.e. assign all time bins in the segment the most frequently
        occurring label in the segment. This transform can only be
        applied if the labelmap contains an 'unlabeled' label,
        because unlabeled segments makes it possible to identify
        the labeled segments. Default is False.
    save_net_outputs : bool
        If True, save 'raw' outputs of neural networks
        before they are converted to annotations. Default is False.
        Typically the output will be "logits"
        to which a softmax transform might be applied.
        For each item in the dataset--each row in  the `dataset_path` .csv--
        the output will be saved in a separate file in `output_dir`,
        with the extension `{MODEL_NAME}.output.npz`. E.g., if the input is a
        spectrogram with `spect_path` filename `gy6or6_032312_081416.npz`,
        and the network is `TweetyNet`, then the net output file
        will be `gy6or6_032312_081416.tweetynet.output.npz`.
    """

    # required, external files
    checkpoint_path = field(converter=expanded_user_path)
    labelmap_path = field(converter=expanded_user_path)

    # required, model / dataloader
    model = field(
        validator=instance_of(ModelConfig),
    )
    batch_size = field(converter=int, validator=instance_of(int))
    dataset: DatasetConfig = field(
        validator=instance_of(DatasetConfig),
    )
    trainer: TrainerConfig = field(
        validator=instance_of(TrainerConfig),
    )

    # optional, transform
    frames_standardizer_path = field(
        converter=converters.optional(expanded_user_path),
        default=None,
    )

    # optional, data loader
    num_workers = field(validator=instance_of(int), default=2)

    annot_csv_filename = field(
        validator=validators.optional(instance_of(str)), default=None
    )
    output_dir = field(
        converter=expanded_user_path,
        default=Path(os.getcwd()),
    )
    min_segment_dur = field(
        validator=validators.optional(instance_of(float)), default=None
    )
    majority_vote = field(validator=instance_of(bool), default=True)
    save_net_outputs = field(validator=instance_of(bool), default=False)

    @classmethod
    def from_config_dict(cls, config_dict: dict) -> PredictConfig:
        """Return :class:`PredictConfig` instance from a :class:`dict`.

        The :class:`dict` passed in should be the one found
        by loading a valid configuration toml file with
        :func:`vak.config.parse.from_toml_path`,
        and then using key ``predict``,
        i.e., ``PredictConfig.from_config_dict(config_dict['predict'])``."""
        for required_key in REQUIRED_KEYS:
            if required_key not in config_dict:
                raise KeyError(
                    "The `[vak.predict]` table in a configuration file requires "
                    f"the option '{required_key}', but it was not found "
                    "when loading the configuration file into a Python dictionary. "
                    "Please check that the configuration file is formatted correctly."
                )
        config_dict["dataset"] = DatasetConfig.from_config_dict(
            config_dict["dataset"]
        )
        config_dict["model"] = ModelConfig.from_config_dict(
            config_dict["model"]
        )
        config_dict["trainer"] = TrainerConfig(**config_dict["trainer"])
        return cls(**config_dict)



# Do this here to suppress warnings before we import vak
import logging
import warnings

from numba.core.errors import NumbaDeprecationWarning
warnings.simplefilter('ignore', category=NumbaDeprecationWarning)

import vak

from . import constants


logger = logging.getLogger(__name__)


def run_prep():
    """Run ``vak prep`` to prepare datasets used with tests.

    This function runs ``prep`` for **only** the configuration files
    in configs.json that have ``null`` for the field in their metadata
    ``'use_dataset_from_config'``. The ``null`` indicates that these
    configs do **not** re-use a dataset prepared from another config.
    """
    configs_to_prep = [
        config_metadata
        for config_metadata in constants.CONFIG_METADATA
        if config_metadata.use_dataset_from_config is None
    ]

    for config_metadata in configs_to_prep:
        config_path = constants.GENERATED_TEST_CONFIGS_ROOT / config_metadata.filename
        if not config_path.exists():
            raise FileNotFoundError(f"{config_path} not found")
        logger.info(
            f"\nRunning vak prep to generate data for tests, using config:\n{config_path.name}"
        )
        vak.cli.prep.prep(toml_path=config_path)



"""A datapipe class used for neural network models with the
frame classification task, where the source data consists of audio signals
or spectrograms of varying lengths."""

from __future__ import annotations

import pathlib
from typing import TYPE_CHECKING

import numpy as np
import numpy.typing as npt
import pandas as pd

from ...transforms.defaults.frame_classification import InferItemTransform
from . import constants, helper
from .metadata import Metadata

if TYPE_CHECKING:
    from ...transforms import FramesStandardizer


class InferDatapipe:
    """A datapipe class used for
    neural network models
    with the frame classification task,
    where the source data consists of audio signals
    or spectrograms of varying lengths.

    Attributes
    ----------
    dataset_path : pathlib.Path
        Path to directory that represents a
        frame classification dataset,
        as created by
        :func:`vak.prep.prep_frame_classification_dataset`.
    split : str
        The name of a split from the dataset,
        one of {'train', 'val', 'test'}.
    subset : str, optional
        Name of subset to use.
        If specified, this takes precedence over split.
        Subsets are typically taken from the training data
        for use when generating a learning curve.
    dataset_df : pandas.DataFrame
        A frame classification dataset,
        represented as a :class:`pandas.DataFrame`.
        This will be only the rows that correspond
        to either ``subset`` or ``split`` from the
        ``dataset_df`` that was passed in when
        instantiating the class.
    frames_paths : numpy.ndarray
        Paths to npy files containing frames,
        either spectrograms or audio signals
        that are input to the model.
    frame_labels_paths : numpy.ndarray
        Paths to npy files containing vectors
        with a label for each frame.
        The targets for the outputs of the model.
    input_type : str
        The type of input to the neural network model.
        One of {'audio', 'spect'}.
    sample_ids : numpy.ndarray
        Indexing vector representing which sample
        from the dataset every frame belongs to.
    inds_in_sample : numpy.ndarray
        Indexing vector representing which index
        within each sample from the dataset
        that every frame belongs to.
    frame_dur: float
        Duration of a frame, i.e., a single sample in audio
        or a single timebin in a spectrogram.
    window_size : int
        Size of windows to return;
        number of frames.
    frames_standardizer : vak.transforms.FramesStandardizer, optional
        Transform applied to frames, the input to the neural network model.
        Optional, default is None.
        If supplied, will be used with the transform applied to inputs and targets,
        :class:`vak.transforms.defaults.frame_classification.TrainItemTransform`.
    """

    def __init__(
        self,
        dataset_path: str | pathlib.Path,
        dataset_df: pd.DataFrame,
        input_type: str,
        split: str,
        sample_ids: npt.NDArray,
        inds_in_sample: npt.NDArray,
        frame_dur: float,
        window_size: int,
        frames_standardizer: FramesStandardizer | None = None,
        frames_padval: float = 0.0,
        frame_labels_padval: int = -1,
        return_padding_mask: bool = False,
        subset: str | None = None,
    ):
        """Initialize a new instance of an :class:`InferDatapipe`.

        Parameters
        ----------
        dataset_path : pathlib.Path
            Path to directory that represents a
            frame classification dataset,
            as created by
            :func:`vak.prep.prep_frame_classification_dataset`.
        dataset_df : pandas.DataFrame
            A frame classification dataset,
            represented as a :class:`pandas.DataFrame`.
        input_type : str
            The type of input to the neural network model.
            One of {'audio', 'spect'}.
        split : str
            The name of a split from the dataset,
            one of {'train', 'val', 'test'}.
        sample_ids : numpy.ndarray
            Indexing vector representing which sample
            from the dataset every frame belongs to.
        inds_in_sample : numpy.ndarray
            Indexing vector representing which index
            within each sample from the dataset
            that every frame belongs to.
        frame_dur: float
            Duration of a frame, i.e., a single sample in audio
            or a single timebin in a spectrogram.
        frames_standardizer : vak.transforms.FramesStandardizer, optional
            Transform applied to frames, the input to the neural network model.
            Optional, default is None.
            If supplied, will be used with the transform applied to inputs and targets,
            :class:`vak.transforms.defaults.frame_classification.InferItemTransform`.
        window_size : int
            Size of windows to return;
            number of frames.
        frames_padval : float
            Value to pad frames with. Added to end of array, the "right side".
            Argument to PadToWindow transform. Default is 0.0.
        frame_labels_padval : int
            Value to pad frame labels vector with. Added to the end of the array.
            Argument to PadToWindow transform. Default is -1.
            Used with ``ignore_index`` argument of :mod:`torch.nn.CrossEntropyLoss`.
        return_padding_mask : bool
            if True, the dictionary returned by ItemTransform classes will include
            a boolean vector to use for cropping back down to size before padding.
            padding_mask has size equal to width of padded array, i.e. original size
            plus padding at the end, and has values of 1 where
            columns in padded are from the original array,
            and values of 0 where columns were added for padding.
        subset : str, optional
            Name of subset to use.
            If specified, this takes precedence over split.
            Subsets are typically taken from the training data
            for use when generating a learning curve.
        """
        from ... import (
            prep,
        )  # avoid circular import, use for constants.INPUT_TYPES

        if input_type not in prep.constants.INPUT_TYPES:
            raise ValueError(
                f"``input_type`` must be one of: {prep.constants.INPUT_TYPES}\n"
                f"Value for ``input_type`` was: {input_type}"
            )

        self.dataset_path = pathlib.Path(dataset_path)
        self.split = split
        self.subset = subset
        # subset takes precedence over split, if specified
        if subset:
            dataset_df = dataset_df[dataset_df.subset == subset].copy()
        else:
            dataset_df = dataset_df[dataset_df.split == split].copy()
        self.dataset_df = dataset_df
        self.input_type = input_type
        self.frames_paths = self.dataset_df[
            constants.FRAMES_PATH_COL_NAME
        ].values
        if split != "predict":
            self.frame_labels_paths = self.dataset_df[
                constants.MULTI_FRAME_LABELS_PATH_COL_NAME
            ].values
        else:
            self.frame_labels_paths = None
        self.sample_ids = sample_ids
        self.inds_in_sample = inds_in_sample
        self.frame_dur = float(frame_dur)
        self.item_transform = InferItemTransform(
            window_size,
            frames_standardizer,
            frames_padval,
            frame_labels_padval,
            return_padding_mask,
        )

    @property
    def duration(self):
        return self.sample_ids.shape[-1] * self.frame_dur

    @property
    def shape(self):
        tmp_x_ind = 0
        tmp_item = self.__getitem__(tmp_x_ind)
        return tmp_item["frames"].shape

    def _load_frames(self, frames_path):
        """Helper function that loads "frames",
        the input to the frame classification model.
        Loads audio or spectrogram, depending on
        :attr:`self.input_type`.
        This function assumes that audio is in wav format
        and spectrograms are in npz files.
        """
        return helper.load_frames(frames_path, self.input_type)

    def __getitem__(self, idx):
        frames_path = self.dataset_path / self.frames_paths[idx]
        frames = self._load_frames(frames_path)
        item = {"frames": frames, "frames_path": frames_path}
        if self.frame_labels_paths is not None:
            frame_labels = np.load(
                self.dataset_path / self.frame_labels_paths[idx]
            )
            item["frame_labels"] = frame_labels

        if self.item_transform:
            item = self.item_transform(**item)

        return item

    def __len__(self):
        """number of batches"""
        return len(np.unique(self.sample_ids))

    @classmethod
    def from_dataset_path(
        cls,
        dataset_path: str | pathlib.Path,
        window_size: int,
        frames_standardizer: FramesStandardizer | None = None,
        frames_padval: float = 0.0,
        frame_labels_padval: int = -1,
        return_padding_mask: bool = False,
        split: str = "val",
        subset: str | None = None,
    ):
        """Make a :class:`InferDatapipe` instance,
        given the path to a frame classification dataset.

        Parameters
        ----------
        dataset_path : pathlib.Path
            Path to directory that represents a
            frame classification dataset,
            as created by
            :func:`vak.prep.prep_frame_classification_dataset`.
        window_size : int
            Size of windows to return;
            number of frames.
        frames_standardizer : vak.transforms.FramesStandardizer, optional
            Transform applied to frames, the input to the neural network model.
            Optional, default is None.
            If supplied, will be used with the transform applied to inputs and targets,
            :class:`vak.transforms.defaults.frame_classification.TrainItemTransform`.
        frames_padval : float
            Value to pad frames with. Added to end of array, the "right side".
            Argument to PadToWindow transform. Default is 0.0.
        frame_labels_padval : int
            Value to pad frame labels vector with. Added to the end of the array.
            Argument to PadToWindow transform. Default is -1.
            Used with ``ignore_index`` argument of :mod:`torch.nn.CrossEntropyLoss`.
        return_padding_mask : bool
            if True, the dictionary returned by ItemTransform classes will include
            a boolean vector to use for cropping back down to size before padding.
            padding_mask has size equal to width of padded array, i.e. original size
            plus padding at the end, and has values of 1 where
            columns in padded are from the original array,
            and values of 0 where columns were added for padding.
        split : str
            The name of a split from the dataset,
            one of {'train', 'val', 'test'}.
            Default is "val".
        subset : str, optional
            Name of subset to use.
            If specified, this takes precedence over split.
            Subsets are typically taken from the training data
            for use when generating a learning curve.

        Returns
        -------
        infer_datapipe : InferDatapipe
        """
        dataset_path = pathlib.Path(dataset_path)
        metadata = Metadata.from_dataset_path(dataset_path)
        frame_dur = metadata.frame_dur
        input_type = metadata.input_type

        dataset_csv_path = dataset_path / metadata.dataset_csv_filename
        dataset_df = pd.read_csv(dataset_csv_path)

        split_path = dataset_path / split
        if subset:
            sample_ids_path = (
                split_path
                / helper.sample_ids_array_filename_for_subset(subset)
            )
        else:
            sample_ids_path = split_path / constants.SAMPLE_IDS_ARRAY_FILENAME
        sample_ids = np.load(sample_ids_path)

        if subset:
            inds_in_sample_path = (
                split_path
                / helper.inds_in_sample_array_filename_for_subset(subset)
            )
        else:
            inds_in_sample_path = (
                split_path / constants.INDS_IN_SAMPLE_ARRAY_FILENAME
            )
        inds_in_sample = np.load(inds_in_sample_path)

        return cls(
            dataset_path,
            dataset_df,
            input_type,
            split,
            sample_ids,
            inds_in_sample,
            frame_dur,
            window_size,
            frames_standardizer,
            frames_padval,
            frame_labels_padval,
            return_padding_mask,
            subset,
        )



"""Tests for vak.train.parametric_umap module"""
import pathlib

import pytest

import vak.config
import vak.common.constants
import vak.common.paths
import vak.train


def assert_train_output_matches_expected(cfg: vak.config.config.Config, model_name: str,
                                         results_path: pathlib.Path):
    model_path = results_path.joinpath(model_name)
    assert model_path.exists()

    tensorboard_log = sorted(
        model_path.glob(f"lightning_logs/**/*events*")
    )
    assert len(tensorboard_log) == 1

    checkpoints_path = model_path.joinpath("checkpoints")
    assert checkpoints_path.exists()
    assert checkpoints_path.joinpath("checkpoint.pt").exists()
    if cfg.train.val_step is not None:
        assert checkpoints_path.joinpath("min-val-loss-checkpoint.pt").exists()


@pytest.mark.slow
@pytest.mark.parametrize(
    "model_name, audio_format, spect_format, annot_format",
    [
        ("ConvEncoderUMAP", "cbin", None, "notmat"),
    ],
)
def test_train_parametric_umap_model(
    model_name, audio_format, spect_format, annot_format,
        specific_config_toml_path, tmp_path, trainer_table
):
    results_path = vak.common.paths.generate_results_dir_name_as_path(tmp_path)
    results_path.mkdir()
    keys_to_change = [
        {"table": "train", "key": "trainer", "value": trainer_table},
        {"table": "train", "key": "root_results_dir", "value": str(results_path)}
    ]
    toml_path = specific_config_toml_path(
        config_type="train",
        model=model_name,
        audio_format=audio_format,
        annot_format=annot_format,
        spect_format=spect_format,
        keys_to_change=keys_to_change,
    )
    cfg = vak.config.Config.from_toml_path(toml_path)

    vak.train.parametric_umap.train_parametric_umap_model(
        model_config=cfg.train.model.asdict(),
        dataset_config=cfg.train.dataset.asdict(),
        trainer_config=cfg.train.trainer.asdict(),
        batch_size=cfg.train.batch_size,
        num_epochs=cfg.train.num_epochs,
        num_workers=cfg.train.num_workers,
        checkpoint_path=cfg.train.checkpoint_path,
        results_path=results_path,
        shuffle=cfg.train.shuffle,
        val_step=cfg.train.val_step,
        ckpt_step=cfg.train.ckpt_step,
    )

    assert_train_output_matches_expected(cfg, cfg.train.model.name, results_path)


@pytest.mark.parametrize(
    'path_option_to_change',
    [
        {"table": "train", "key": "checkpoint_path", "value": '/obviously/doesnt/exist/ckpt.pt'},
    ]
)
def test_train_parametric_umap_model_raises_file_not_found(
    path_option_to_change, specific_config_toml_path, tmp_path, trainer_table
):
    """Test that pre-conditions in :func:`vak.train.parametric_umap.train_parametric_umap_model`
    raise FileNotFoundError when one of the following does not exist:
    checkpoint_path, dataset_path
    """
    keys_to_change = [
        {"table": "train", "key": "trainer", "value": trainer_table},
        path_option_to_change
    ]
    toml_path = specific_config_toml_path(
        config_type="train",
        model="ConvEncoderUMAP",
        audio_format="cbin",
        annot_format="notmat",
        spect_format=None,
        keys_to_change=keys_to_change,
    )
    cfg = vak.config.Config.from_toml_path(toml_path)
    results_path = vak.common.paths.generate_results_dir_name_as_path(tmp_path)
    results_path.mkdir()

    with pytest.raises(FileNotFoundError):
        vak.train.parametric_umap.train_parametric_umap_model(
            model_config=cfg.train.model.asdict(),
            dataset_config=cfg.train.dataset.asdict(),
            trainer_config=cfg.train.trainer.asdict(),
            batch_size=cfg.train.batch_size,
            num_epochs=cfg.train.num_epochs,
            num_workers=cfg.train.num_workers,
            checkpoint_path=cfg.train.checkpoint_path,
            results_path=results_path,
            shuffle=cfg.train.shuffle,
            val_step=cfg.train.val_step,
            ckpt_step=cfg.train.ckpt_step,
        )


@pytest.mark.parametrize(
    'path_option_to_change',
    [
        {"table": "train", "key": ["dataset", "path"], "value": '/obviously/doesnt/exist/dataset-dir'},
        {"table": "train", "key": "root_results_dir", "value": '/obviously/doesnt/exist/results/'},
    ]
)
def test_train_parametric_umap_model_raises_not_a_directory(
    path_option_to_change, specific_config_toml_path, trainer_table, tmp_path
):
    """Test that core.train raises NotADirectory
    when directory does not exist
    """
    keys_to_change = [
        path_option_to_change,
        {"table": "train", "key": "trainer", "value": trainer_table},
    ]

    toml_path = specific_config_toml_path(
        config_type="train",
        model="ConvEncoderUMAP",
        audio_format="cbin",
        annot_format="notmat",
        spect_format=None,
        keys_to_change=keys_to_change,
    )
    cfg = vak.config.Config.from_toml_path(toml_path)
    model_config = cfg.train.model.asdict()

    # mock behavior of cli.train, building `results_path` from config option `root_results_dir`
    results_path = cfg.train.root_results_dir / 'results-dir-timestamp'

    with pytest.raises(NotADirectoryError):
        vak.train.parametric_umap.train_parametric_umap_model(
            model_config=model_config,
            dataset_config=cfg.train.dataset.asdict(),
            trainer_config=cfg.train.trainer.asdict(),
            batch_size=cfg.train.batch_size,
            num_epochs=cfg.train.num_epochs,
            num_workers=cfg.train.num_workers,
            checkpoint_path=cfg.train.checkpoint_path,
            results_path=results_path,
            shuffle=cfg.train.shuffle,
            val_step=cfg.train.val_step,
            ckpt_step=cfg.train.ckpt_step,
        )



"""Functions for creating splits of datasets used with neural network models,
such as the standard train-val-test splits used with supervised learning methods."""

from __future__ import annotations

import logging
import pathlib

import numpy as np
import pandas as pd

from ...common.labels import from_df as labels_from_df
from .algorithms import brute_force
from .algorithms.validate import validate_split_durations

logger = logging.getLogger(__name__)


def train_test_dur_split_inds(
    durs,
    labels,
    labelset,
    train_dur,
    test_dur,
    val_dur=None,
    algo="brute_force",
):
    """Return indices to split a dataset into training, test, and validation sets of specified durations.

    Given the durations of a set of vocalizations, and labels from the annotations for those vocalizations,
    this function returns arrays of indices for splitting up the set into training, test,
    and validation sets.

    Using those indices will produce datasets that each contain instances of all labels in the set of labels.

    Parameters
    ----------
    durs : iterable
        Of float. Durations of audio files.
    labels : iterable
        Of numpy arrays of str or int. Labels for segments (syllables, phonemes, etc.) in audio files.
    labelset : set, list
        set of unique labels for segments in files. Used to verify that each returned array
        of indices will produce a set that contains instances of all labels found in original
        set.
    train_dur : float
        Target duration for training set, in seconds.
    test_dur : float
        Target duration for test set, in seconds.
    val_dur : float
        Target duration for validation set, in seconds. Default is None.
        If None, no indices are returned for validation set.
    algo : str
        algorithm to use. One of {'brute_force', 'inc_freq'}. Default is 'brute_force'. For more information
        on the algorithms, see the docstrings, e.g., vak.io.algorithms.brute_force

    Returns
    -------
    train_inds, test_inds, val_inds : numpy.ndarray
        indices to use with some array-like object to produce sets of specified durations
    """
    if len(durs) != len(labels):
        raise ValueError(
            f"length of durs, {len(durs)} does not equal length of labels, {len(labels)}"
        )

    total_dur = sum(durs)
    train_dur, val_dur, test_dur = validate_split_durations(
        train_dur, val_dur, test_dur, total_dur
    )

    total_target_dur = sum(
        [dur for dur in (train_dur, test_dur, val_dur) if dur is not None]
    )

    if total_target_dur > total_dur:
        raise ValueError(
            f"Total duration of dataset, {total_dur} seconds, is less than total target duration of "
            f"training, test, and (if specified) validation sets: {total_target_dur}"
        )

    logger.info(
        f"Total target duration of splits: {total_target_dur} seconds. "
        f"Will be drawn from dataset with total duration: {total_dur:.3f}.",
    )

    if algo == "brute_force":
        train_inds, val_inds, test_inds = brute_force(
            durs, labels, labelset, train_dur, val_dur, test_dur
        )
    else:
        raise NotImplementedError(f"algorithm {algo} not implemented")

    return train_inds, val_inds, test_inds


def frame_classification_dataframe(
    dataset_df: pd.DataFrame,
    dataset_path: str | pathlib.Path,
    labelset: set,
    train_dur: float | None = None,
    test_dur: float | None = None,
    val_dur: float | None = None,
):
    """Create datasets splits from a dataframe
    representing a frame classification dataset.

    Splits dataset into training, test, and (optionally) validation subsets,
    specified by their duration.

    Additionally, adds a 'split' column to the dataframe,
    that assigns each row to 'train', 'val', 'test', or 'None'.

    Parameters
    ----------
    dataset_df : pandas.Dataframe
        A pandas DataFrame representing the samples in a dataset generated by ``vak prep``.
    dataset_path : str
        Path to dataset, a directory generated by running ``vak prep``.
    labelset : set, list
        The set of label classes for vocalizations in dataset.
    train_dur : float
        Total duration of training set, in seconds. Default is None
    test_dur : float
        Total duration of test set, in seconds. Default is None.
    val_dur : float
        Total duration of validation set, in seconds. Default is None.

    Returns
    -------
    dataset_df : pandas.Dataframe
        A copy of the input dataset with a 'split' column added,
        that assigns each vocalization (row) to a subset,
        i.e., train, validation, or test.
        If the vocalization was not added to one of the subsets,
        its value for 'split' will be 'None'.

    Notes
    -----
    Uses the function :func:`vak.dataset.split.train_test_dur_split_inds`
    to find indices for each subset.
    """
    dataset_path = pathlib.Path(dataset_path)
    if not dataset_path.exists() or not dataset_path.is_dir():
        raise NotADirectoryError(
            f"`dataset_path` not found or not recognized as a directory: {dataset_path}"
        )

    dataset_df = (
        dataset_df.copy()
    )  # don't want this function to have unexpected side effects, so return a copy
    labels = labels_from_df(dataset_df, dataset_path)

    durs = dataset_df["duration"].values
    train_inds, val_inds, test_inds = train_test_dur_split_inds(
        durs=durs,
        labels=labels,
        labelset=labelset,
        train_dur=train_dur,
        test_dur=test_dur,
        val_dur=val_dur,
    )

    # start off with all elements set to 'None'
    # so we don't have to change any that are not assigned to one of the subsets to 'None' after
    split_col = np.asarray(
        ["None" for _ in range(len(dataset_df))], dtype="object"
    )
    split_zip = zip(
        ["train", "val", "test"], [train_inds, val_inds, test_inds]
    )
    for split_name, split_inds in split_zip:
        if split_inds is not None:
            split_col[split_inds] = split_name

    # add split column to dataframe
    dataset_df["split"] = split_col

    return dataset_df


def unit_dataframe(
    dataset_df: pd.DataFrame,
    dataset_path: str | pathlib.Path,
    labelset: set,
    train_dur: float | None = None,
    test_dur: float | None = None,
    val_dur: float | None = None,
):
    """Create datasets splits from a dataframe
    representing a unit dataset.

    Splits dataset into training, test, and (optionally) validation subsets,
    specified by their duration.

    Additionally adds a 'split' column to the dataframe,
    that assigns each row to 'train', 'val', 'test', or 'None'.

    Parameters
    ----------
    dataset_df : pandas.Dataframe
        A pandas DataFrame representing the samples in a dataset,
        generated by ``vak prep``.
    dataset_path : str
        Path to dataset, a directory generated by running ``vak prep``.
    labelset : set, list
        The set of label classes for vocalizations in dataset.
    train_dur : float
        Total duration of training set, in seconds. Default is None
    test_dur : float
        Total duration of test set, in seconds. Default is None.
    val_dur : float
        Total duration of validation set, in seconds. Default is None.

    Returns
    -------
    dataset_df : pandas.Dataframe
        A copy of the input dataset with a 'split' column added,
        that assigns each vocalization (row) to a subset,
        i.e., train, validation, or test.
        If the vocalization was not added to one of the subsets,
        its value for 'split' will be 'None'.

    Notes
    -----
    Uses the function :func:`vak.dataset.split.train_test_dur_split_inds`
    to find indices for each subset.
    """
    dataset_path = pathlib.Path(dataset_path)
    if not dataset_path.exists() or not dataset_path.is_dir():
        raise NotADirectoryError(
            f"`dataset_path` not found or not recognized as a directory: {dataset_path}"
        )

    dataset_df = (
        dataset_df.copy()
    )  # don't want this function to have unexpected side effects, so return a copy
    labels = [np.array([label]) for label in dataset_df.label.values]

    durs = dataset_df["duration"].values
    train_inds, val_inds, test_inds = train_test_dur_split_inds(
        durs=durs,
        labels=labels,
        labelset=labelset,
        train_dur=train_dur,
        test_dur=test_dur,
        val_dur=val_dur,
    )

    # start off with all elements set to 'None'
    # so we don't have to change any that are not assigned to one of the subsets to 'None' after
    split_col = np.asarray(
        ["None" for _ in range(len(dataset_df))], dtype="object"
    )
    split_zip = zip(
        ["train", "val", "test"], [train_inds, val_inds, test_inds]
    )
    for split_name, split_inds in split_zip:
        if split_inds is not None:
            split_col[split_inds] = split_name

    # add split column to dataframe
    dataset_df["split"] = split_col

    return dataset_df



import inspect

import torch
import pytest

import vak.nets


class TestConvEncoder:

    @pytest.mark.parametrize(
        'input_shape',
        [
            (
                    1, 128, 128,
            ),
            (
                    1, 256, 256,
            ),
        ]
    )
    def test_init(self, input_shape):
        """test we can instantiate ConvEncoder
        and it has the expected attributes"""
        net = vak.nets.ConvEncoder(input_shape)
        assert isinstance(net, vak.nets.ConvEncoder)
        for expected_attr, expected_type in (
            ('input_shape', tuple),
            ('num_input_channels', int),
            ('conv', torch.nn.Module),
            ('encoder', torch.nn.Module),
        ):
            assert hasattr(net, expected_attr)
            assert isinstance(getattr(net, expected_attr), expected_type)

        assert net.input_shape == input_shape

    @pytest.mark.parametrize(
        'input_shape, batch_size',
        [
            ((1, 128, 128,), 32),
            ((1, 256, 256,), 64),
        ]
    )
    def test_forward(self, input_shape, batch_size):
        """test we can forward a tensor through a ConvEncoder instance
        and get the expected output"""

        input = torch.rand(batch_size, *input_shape)  # a "batch"
        net = vak.nets.ConvEncoder(input_shape)
        out = net(input)
        assert isinstance(out, torch.Tensor)




import json
import pathlib

import pytest

import vak.datapipes.frame_classification


ARGNAMES = 'dataset_csv_filename, input_type, frame_dur'
ARGVALS = [
    (pathlib.Path('bird1_prep_230319_115852.csv'), 'spect', 0.002),
    (pathlib.Path('bird1_prep_230319_115852.csv'), 'spect', 0.001),
    (pathlib.Path('bird1_prep_230319_115852.csv'), 'spect', 0.0027),
    (pathlib.Path('bird1_prep_230319_115852.csv'), 'audio', 3.125e-05),
    (pathlib.Path('bird1_prep_230319_115852.csv'), 'audio', 2.2727272727272726e-05),
    (pathlib.Path('bird1_prep_230319_115852.csv'), 'audio', 6.25e-05),
]


class TestMetadata:
    @pytest.mark.parametrize(
        ARGNAMES,
        ARGVALS
    )
    def test_metadata_init(self, dataset_csv_filename, input_type, frame_dur):
        metadata = vak.datapipes.frame_classification.Metadata(dataset_csv_filename, input_type, frame_dur)
        assert isinstance(metadata, vak.datapipes.frame_classification.Metadata)
        for attr_name, attr_val in zip(
            ('dataset_csv_filename', 'input_type', 'frame_dur'),
            (dataset_csv_filename, input_type, frame_dur),
        ):
            assert hasattr(metadata, attr_name)
            if isinstance(attr_val, pathlib.Path):
                assert getattr(metadata, attr_name) == str(attr_val)
            else:
                assert getattr(metadata, attr_name) == attr_val

    @pytest.mark.parametrize(
        ARGNAMES,
        ARGVALS
    )
    def test_metadata_from_path(self, dataset_csv_filename, input_type, frame_dur, tmp_path):
        # we make metadata "by hand"
        metadata_dict = {
            'dataset_csv_filename': str(dataset_csv_filename),
            'input_type': input_type,
            'frame_dur': frame_dur,
        }
        metadata_json_path = tmp_path / vak.datapipes.frame_classification.Metadata.METADATA_JSON_FILENAME
        with metadata_json_path.open('w') as fp:
            json.dump(metadata_dict, fp, indent=4)

        metadata = vak.datapipes.frame_classification.Metadata.from_path(metadata_json_path)
        assert isinstance(metadata, vak.datapipes.frame_classification.Metadata)
        for attr_name, attr_val in zip(
            ('dataset_csv_filename', 'input_type', 'frame_dur'),
            (dataset_csv_filename, input_type, frame_dur),
        ):
            assert hasattr(metadata, attr_name)
            if isinstance(attr_val, pathlib.Path):
                assert getattr(metadata, attr_name) == str(attr_val)
            else:
                assert getattr(metadata, attr_name) == attr_val

    @pytest.mark.parametrize(
        ARGNAMES,
        ARGVALS
    )
    def test_metadata_to_json(self, dataset_csv_filename, input_type, frame_dur, tmp_path):
        metadata_to_json = vak.datapipes.frame_classification.Metadata(dataset_csv_filename, input_type, frame_dur)
        mock_dataset_path = tmp_path / 'mock_dataset'
        mock_dataset_path.mkdir()

        metadata_to_json.to_json(dataset_path=mock_dataset_path)
        expected_json_path = mock_dataset_path / vak.datapipes.frame_classification.Metadata.METADATA_JSON_FILENAME
        assert expected_json_path.exists()

        metadata_from_json = vak.datapipes.frame_classification.Metadata.from_path(expected_json_path)
        assert metadata_from_json == metadata_to_json



from __future__ import annotations

import logging
import os
from pathlib import Path

import dask.bag as db
import numpy as np
from dask.diagnostics import ProgressBar

from ... import config
from ...common import constants, files
from ...common.annotation import map_annotated_to_annot
from ...common.converters import labelset_to_set
from ...config.spect_params import SpectParamsConfig
from .spect import spectrogram

logger = logging.getLogger(__name__)


def files_from_dir(audio_dir, audio_format):
    """get all audio files of a given format
    from a directory or its sub-directories,
    using the file extension associated with that annotation format.

    Parameters
    ----------
    audio_dir : str
        path to directory containing audio files.
    audio_format : str
        valid audio file format. One of {'wav', 'cbin'}.

    Returns
    -------
    audio_files : list
        of paths to audio files
    """
    if audio_format not in constants.AUDIO_FORMAT_FUNC_MAP:
        raise ValueError(f"'{audio_format}' is not a valid audio format")
    audio_files = files.from_dir(audio_dir, audio_format)
    return audio_files


def make_spectrogram_files_from_audio_files(
    audio_format: str,
    spect_params: dict | config.spect_params.SpectParamsConfig,
    output_dir: str,
    audio_dir: list | None = None,
    audio_files: list | None = None,
    annot_list: list | None = None,
    audio_annot_map: dict | None = None,
    annot_format: str | None = None,
    labelset: str | list | None = None,
    dask_bag_kwargs: dict | None = None,
):
    """Make spectrograms from audio files and save them in npz array files.

    Parameters
    ----------
    audio_format : str
        A :class:`string` representing the format of audio files.
        One of :constant:`vak.common.constants.VALID_AUDIO_FORMATS`.
    spect_params : dict or config.spect_params.SpectParamsConfig
        parameters for computing spectrogram, from .toml file.
        To see all related parameters, run:
        >>> help(vak.config.spect_params.SpectParamsConfig)
        To get a default configuration, create a SpectParamConfig
        with no arguments and then pass that to `to_spect`:
        >>> default_spect_params = vak.config.spect_params.SpectParamsConfig()
        >>> to_spect(audio_format='wav', spect_params=default_spect_params, output_dir='.')
    audio_dir : str
        Path to directory containing audio files from which to make spectrograms.
    audio_files : list
        of str, full paths to audio files from which to make spectrograms
    annot_list : list
        of annotations for array files. Default is None.
    audio_annot_map : dict
        Where keys are paths to array files and value corresponding to each key is
        the annotation for that array file.
        Default is None.
    output_dir : str
        directory in which to save .spect.npz file generated for each audio file.
    labelset : str, list
        of str or int, set of unique labels for vocalizations. Default is None.
        If not None, skip files where the associated annotations contain labels not in ``labelset``.
        ``labelset`` is converted to a Python ``set`` using ``vak.converters.labelset_to_set``.
        See help for that function for details on how to specify labelset.
    dask_bag_kwargs : dict
        Keyword arguments used when calling ``dask.bag.from_sequence``.
        E.g., ``{npartitions=20}``. Allows for finer-grained control
        when needed to process files of different sizes.

    Returns
    -------
    spect_files : list
        of str, full paths to .spect.npz files

    Notes
    -----
    For each audio file, a corresponding 'spect.npz' file will be created.
    Each '.spect.npz' file contains the following arrays:
        s : numpy.ndarray
            spectrogram, a 2-d array
        f : numpy.ndarray
            vector of centers of frequency bins from spectrogram
        t : numpy.ndarray
            vector of centers of tme bins from spectrogram
        audio_path : numpy.ndarray
            path to source audio file used to create spectrogram

    The names of the arrays are defaults, and will change if different values are specified
    in spect_params for 'spect_key', 'freqbins_key', 'timebins_key', or 'audio_path_key'.
    """
    if audio_format not in constants.VALID_AUDIO_FORMATS:
        raise ValueError(
            f"audio format must be one of '{constants.VALID_AUDIO_FORMATS}'; "
            f"format '{audio_format}' not recognized."
        )

    if all([arg is None for arg in (audio_dir, audio_files, audio_annot_map)]):
        raise ValueError(
            "must specify one of: audio_dir, audio_files, audio_annot_map"
        )

    if audio_dir and audio_files:
        raise ValueError(
            "received values for audio_dir and audio_files, unclear which to use"
        )

    if audio_dir and audio_annot_map:
        raise ValueError(
            "received values for audio_dir and audio_annot_map, unclear which to use"
        )

    if audio_files and audio_annot_map:
        raise ValueError(
            "received values for audio_files and audio_annot_map, unclear which to use"
        )

    if annot_list and audio_annot_map:
        raise ValueError(
            "received values for annot_list and array_annot_map, unclear which annotations to use"
        )

    if labelset is not None:
        labelset = labelset_to_set(labelset)

    if type(spect_params) not in [dict, SpectParamsConfig]:
        raise TypeError(
            "type of spect_params must be an instance of vak.config.spect_params.SpectParamsConfig, "
            "or a dict that can be converted to a SpectParamsConfig instance, "
            f"but was {type(spect_params)}"
        )
    if type(spect_params) is dict:
        spect_params = SpectParamsConfig(**spect_params)

    if dask_bag_kwargs is None:
        dask_bag_kwargs = {}  # so ``db.bag(**dask_bag_kwargs)`` works below

    # validate audio files if supplied by user
    if audio_files:
        # make sure audio files are all the same type, and the same as audio format specified
        exts = []
        for audio_file in audio_files:
            root, ext = os.path.splitext(audio_file)
            ext = (
                ext.lower()
            )  # make case-insensitive, e.g. treat .wav and .WAV the same
            exts.append(ext)
        uniq_ext = set(exts)
        if len(uniq_ext) > 1:
            raise ValueError(
                "audio_files should all have the same extension, "
                f"but found more than one: {uniq_ext}"
            )
        else:
            ext_str = uniq_ext.pop()
            if audio_format not in ext_str:
                raise ValueError(
                    f"audio format. '{audio_format}', does not match extensions in audio_files, '{ext_str}''"
                )

    # otherwise get audio files using audio dir (won't need to validate audio files)
    if audio_dir:
        audio_files = files_from_dir(audio_dir, audio_format)

    if annot_list:
        audio_annot_map = map_annotated_to_annot(
            audio_files, annot_list, annot_format
        )

    logger.info("creating array files with spectrograms")

    # use mapping (if generated/supplied) with labelset, if supplied, to filter
    if audio_annot_map:
        if labelset:  # then remove annotations with labels not in labelset
            # note we do this here so it happens regardless of whether
            # user supplied audio_annot_map *or* we constructed it above
            for audio_file, annot in list(audio_annot_map.items()):
                # loop in a verbose way (i.e. not a comprehension)
                # so we can give user warning when we skip files
                annot_labelset = set(annot.seq.labels)
                # below, set(labels_mapping) is a set of that dict's keys
                if not annot_labelset.issubset(set(labelset)):
                    # because there's some label in labels that's not in labelset
                    audio_annot_map.pop(audio_file)
                    extra_labels = annot_labelset - labelset
                    logger.info(
                        f"Found labels, {extra_labels}, in {Path(audio_file).name}, "
                        "that are not in labels_mapping. Skipping file.",
                    )
        audio_files = sorted(list(audio_annot_map.keys()))

    # this is defined here so all other arguments to 'to_spect' are in scope
    def _spect_file(audio_file):
        """helper function that enables parallelized creation of array
        files containing spectrograms.
        Accepts path to audio file, saves .npz file with spectrogram"""
        dat, fs = constants.AUDIO_FORMAT_FUNC_MAP[audio_format](audio_file)
        s, f, t = spectrogram(
            dat,
            fs,
            spect_params.fft_size,
            spect_params.step_size,
            spect_params.thresh,
            spect_params.transform_type,
            spect_params.freq_cutoffs,
        )
        spect_dict = {
            spect_params.spect_key: s,
            spect_params.freqbins_key: f,
            spect_params.timebins_key: t,
            # cast audio_file to str so that we don't
            # end up with a pickled pathlib.Path in .npz file
            spect_params.audio_path_key: str(audio_file),
        }
        basename = os.path.basename(audio_file)
        npz_fname = os.path.join(
            os.path.normpath(output_dir),
            basename + constants.SPECT_NPZ_EXTENSION,
        )
        np.savez(npz_fname, **spect_dict)
        return npz_fname

    bag = db.from_sequence(audio_files, **dask_bag_kwargs)
    with ProgressBar():
        spect_files = list(bag.map(_spect_file))
    # sort because ordering from Dask not guaranteed
    spect_files = sorted(spect_files)
    return spect_files



"""TweetyNet model"""

from __future__ import annotations

import torch
from torch import nn

from ..nn.modules import Conv2dTF


class TweetyNet(nn.Module):
    """Neural network architecture
    that assign labels to time bins
    ("frames") in spectrogram windows.

    as described in
    https://elifesciences.org/articles/63853
    https://github.com/yardencsGitHub/tweetynet

    Cohen, Y., Nicholson, D. A., Sanchioni, A., Mallaber, E. K., Skidanova, V., & Gardner, T. J. (2022).
    Automated annotation of birdsong with a neural network that segments spectrograms. Elife, 11, e63853.

    Attributes
    ----------
    num_classes : int
        Number of classes.
        One of the two dimensions of the output.
    input_shape : tuple(int)
        With dimensions
        (channels, num. frequency bins, num. time bins in window).
    cnn : torch.nn.Sequential
        Convolutional layers of model.
    rnn_input_size : int
        Size of input to TweetyNet.rnn.
        Will be the product of the first two dimensions
        of the output of ``TweetyNet.cnn``,
        i.e. the number of output channels times
        the number of elements in the dimension
        that corresponds to frequency bins in the input.
    rnn : torch.nn.LSTM
        Bidirectional LSTM layer,
        that receives output of ``TweetyNet.cnn``.
    fc : torch.nn.Linear
        Finally fully-connected layer that maps
        the output of ``TweetyNet.rnn`` to a
        matrix of size (num. time bins in window, num. classes).

    Notes
    -----
    This is the network used by ``vak.models.TweetyNetModel``.
    """

    def __init__(
        self,
        num_classes,
        num_input_channels=1,
        num_freqbins=256,
        padding="SAME",
        conv1_filters=32,
        conv1_kernel_size=(5, 5),
        conv2_filters=64,
        conv2_kernel_size=(5, 5),
        pool1_size=(8, 1),
        pool1_stride=(8, 1),
        pool2_size=(8, 1),
        pool2_stride=(8, 1),
        hidden_size=None,
        rnn_dropout=0.0,
        num_layers=1,
        bidirectional=True,
    ):
        """initialize TweetyNet model

        Parameters
        ----------
        num_classes : int
            Number of classes to predict, e.g., number of syllable classes in an individual bird's song
        num_input_channels: int
            Number of channels in input. Typically one, for a spectrogram.
            Default is 1.
        num_freqbins: int
            Number of frequency bins in spectrograms that will be input to model.
            Default is 256.
        padding : str
            type of padding to use, one of {"VALID", "SAME"}. Default is "SAME".
        conv1_filters : int
            Number of filters in first convolutional layer. Default is 32.
        conv1_kernel_size : tuple
            Size of kernels, i.e. filters, in first convolutional layer. Default is (5, 5).
        conv2_filters : int
            Number of filters in second convolutional layer. Default is 64.
        conv2_kernel_size : tuple
            Size of kernels, i.e. filters, in second convolutional layer. Default is (5, 5).
        pool1_size : two element tuple of ints
            Size of sliding window for first max pooling layer. Default is (1, 8)
        pool1_stride : two element tuple of ints
            Step size for sliding window of first max pooling layer. Default is (1, 8)
        pool2_size : two element tuple of ints
            Size of sliding window for second max pooling layer. Default is (1, 8),
        pool2_stride : two element tuple of ints
            Step size for sliding window of second max pooling layer. Default is (1, 8)
        hidden_size : int
            number of features in the hidden state ``h``. Default is None,
            in which case ``hidden_size`` is set to the dimensionality of the
            output of the convolutional neural network. This default maintains
            the original behavior of the network.
        rnn_dropout : float
            If non-zero, introduces a Dropout layer on the outputs of each LSTM layer except the last layer,
            with dropout probability equal to dropout. Default: 0
        num_layers : int
            Number of recurrent layers. Default is 1.
        bidirectional : bool
            If True, make LSTM bidirectional. Default is True.
        """
        super().__init__()
        self.num_classes = num_classes
        self.num_input_channels = num_input_channels
        self.num_freqbins = num_freqbins

        self.cnn = nn.Sequential(
            Conv2dTF(
                in_channels=self.num_input_channels,
                out_channels=conv1_filters,
                kernel_size=conv1_kernel_size,
                padding=padding,
            ),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=pool1_size, stride=pool1_stride),
            Conv2dTF(
                in_channels=conv1_filters,
                out_channels=conv2_filters,
                kernel_size=conv2_kernel_size,
                padding=padding,
            ),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=pool2_size, stride=pool2_stride),
        )

        # determine number of features in output after stacking channels
        # we use the same number of features for hidden states
        # note self.num_hidden is also used to reshape output of cnn in self.forward method
        N_DUMMY_TIMEBINS = (
            256  # some not-small number. This dimension doesn't matter here
        )
        batch_shape = (
            1,
            self.num_input_channels,
            self.num_freqbins,
            N_DUMMY_TIMEBINS,
        )
        tmp_tensor = torch.rand(batch_shape)
        tmp_out = self.cnn(tmp_tensor)
        channels_out, freqbins_out = tmp_out.shape[1], tmp_out.shape[2]
        self.rnn_input_size = channels_out * freqbins_out

        if hidden_size is None:
            self.hidden_size = self.rnn_input_size
        else:
            self.hidden_size = hidden_size

        self.rnn = nn.LSTM(
            input_size=self.rnn_input_size,
            hidden_size=self.hidden_size,
            num_layers=num_layers,
            dropout=rnn_dropout,
            bidirectional=bidirectional,
        )

        # for self.fc, in_features = hidden_size * 2 because LSTM is bidirectional
        # so we get hidden forward + hidden backward as output
        self.fc = nn.Linear(
            in_features=self.hidden_size * 2, out_features=num_classes
        )

    def forward(self, x):
        features = self.cnn(x)
        # stack channels, to give tensor shape (batch, rnn_input_size, num time bins)
        features = features.view(features.shape[0], self.rnn_input_size, -1)
        # switch dimensions for feeding to rnn, to (num time bins, batch size, input size)
        features = features.permute(2, 0, 1)
        rnn_output, _ = self.rnn(features)
        # permute back to (batch, time bins, hidden size) to project features down onto number of classes
        rnn_output = rnn_output.permute(1, 0, 2)
        logits = self.fc(rnn_output)
        # permute yet again so that dimension order is (batch, classes, time steps)
        # because this is order that loss function expects
        return logits.permute(0, 2, 1)



"""Helper functions that return names of directories
generated during a run of a learning curve."""


def train_dur_dirname(train_dur: float) -> str:
    """Returns name of directory for all replicates
    trained with a training set of a specified duration,
    ``f"train_dur_{train_dur}s"``.
    """
    return f"train_dur_{float(train_dur)}s"


def replicate_dirname(replicate_num: int) -> str:
    """Returns name of directory for a replicate,
    ``f"replicate_{replicate_num}``.
    """
    return f"replicate_{int(replicate_num)}"



"""tests for vak.learncurve.frame_classification module"""
import pytest

import vak.config
import vak.common.constants
import vak.learncurve
import vak.common.paths


def assert_learncurve_output_matches_expected(cfg, model_name, results_path):
    assert results_path.joinpath("learning_curve.csv").exists()

    for train_set_dur in cfg.prep.train_set_durs:
        train_set_dur_root = results_path / vak.learncurve.dirname.train_dur_dirname(train_set_dur)
        assert train_set_dur_root.exists()

        for replicate_num in range(1, cfg.prep.num_replicates + 1):
            replicate_path = train_set_dur_root / vak.learncurve.dirname.replicate_dirname(replicate_num)
            assert replicate_path.exists()

            assert replicate_path.joinpath("labelmap.json").exists()

            if cfg.learncurve.standardize_frames:
                assert replicate_path.joinpath("FramesStandardizer").exists()

            eval_csv = sorted(replicate_path.glob(f"eval_{model_name}*csv"))
            assert len(eval_csv) == 1

            model_path = replicate_path.joinpath(model_name)
            assert model_path.exists()

            tensorboard_log = sorted(
                replicate_path.glob(f"lightning_logs/**/*events*")
            )
            assert len(tensorboard_log) == 1

            checkpoints_path = model_path.joinpath("checkpoints")
            assert checkpoints_path.exists()
            assert checkpoints_path.joinpath("checkpoint.pt").exists()
            if cfg.learncurve.val_step is not None:
                assert checkpoints_path.joinpath(
                    "max-val-acc-checkpoint.pt"
                ).exists()


@pytest.mark.slow
@pytest.mark.parametrize(
    'model_name, audio_format, annot_format',
    [
        ("TweetyNet", "cbin", "notmat"),
    ]
)
def test_learning_curve_for_frame_classification_model(
        model_name, audio_format, annot_format, specific_config_toml_path, tmp_path, trainer_table):
    keys_to_change = {"table": "learncurve", "key": "trainer", "value": trainer_table}

    toml_path = specific_config_toml_path(
        config_type="learncurve",
        model=model_name,
        audio_format=audio_format,
        annot_format=annot_format,
        keys_to_change=keys_to_change,
    )

    cfg = vak.config.Config.from_toml_path(toml_path)
    results_path = vak.common.paths.generate_results_dir_name_as_path(tmp_path)
    results_path.mkdir()

    vak.learncurve.frame_classification.learning_curve_for_frame_classification_model(
        model_config=cfg.learncurve.model.asdict(),
        dataset_config=cfg.learncurve.dataset.asdict(),
        trainer_config=cfg.learncurve.trainer.asdict(),
        batch_size=cfg.learncurve.batch_size,
        num_epochs=cfg.learncurve.num_epochs,
        num_workers=cfg.learncurve.num_workers,
        results_path=results_path,
        post_tfm_kwargs=cfg.learncurve.post_tfm_kwargs,
        standardize_frames=cfg.learncurve.standardize_frames,
        shuffle=cfg.learncurve.shuffle,
        val_step=cfg.learncurve.val_step,
        ckpt_step=cfg.learncurve.ckpt_step,
        patience=cfg.learncurve.patience,
    )

    assert_learncurve_output_matches_expected(cfg, cfg.learncurve.model.name, results_path)


@pytest.mark.parametrize(
    'dir_option_to_change',
    [
        {"table": "learncurve", "key": "root_results_dir", "value": '/obviously/does/not/exist/results/'},
        {"table": "learncurve", "key": ["dataset", "path"], "value": '/obviously/doesnt/exist/dataset-dir'},
    ]
)
def test_learncurve_raises_not_a_directory(dir_option_to_change,
                                           specific_config_toml_path,
                                           tmp_path, trainer_table):
    """Test that core.learncurve.learning_curve raises NotADirectoryError
    when the following directories do not exist:
    results_path, previous_run_path, dataset_path
    """
    keys_to_change = [
        {"table": "learncurve", "key": "trainer", "value": trainer_table},
        dir_option_to_change
    ]
    toml_path = specific_config_toml_path(
        config_type="learncurve",
        model="TweetyNet",
        audio_format="cbin",
        annot_format="notmat",
        keys_to_change=keys_to_change,
    )
    cfg = vak.config.Config.from_toml_path(toml_path)
    # mock behavior of cli.learncurve, building `results_path` from config option `root_results_dir`
    results_path = cfg.learncurve.root_results_dir / 'results-dir-timestamp'

    with pytest.raises(NotADirectoryError):
        vak.learncurve.frame_classification.learning_curve_for_frame_classification_model(
            model_config=cfg.learncurve.model.asdict(),
            dataset_config=cfg.learncurve.dataset.asdict(),
            trainer_config=cfg.learncurve.trainer.asdict(),
            batch_size=cfg.learncurve.batch_size,
            num_epochs=cfg.learncurve.num_epochs,
            num_workers=cfg.learncurve.num_workers,
            results_path=results_path,
            post_tfm_kwargs=cfg.learncurve.post_tfm_kwargs,
            standardize_frames=cfg.learncurve.standardize_frames,
            shuffle=cfg.learncurve.shuffle,
            val_step=cfg.learncurve.val_step,
            ckpt_step=cfg.learncurve.ckpt_step,
            patience=cfg.learncurve.patience,
        )



from . import frame_classification, learncurve, make_splits, validators
from .assign_samples_to_splits import assign_samples_to_splits
from .frame_classification import prep_frame_classification_dataset
from .source_files import get_or_make_source_files

__all__ = [
    "assign_samples_to_splits",
    "frame_classification",
    "get_or_make_source_files",
    "learncurve",
    "make_splits",
    "prep_frame_classification_dataset",
    "validators",
]



"""tests for functional forms of transforms
for labeled timebins.

Tests are in the same order as the module ``vak.transforms.frame_labels.functional``.:
- from_segments: transform to get labeled timebins from annotations
- to_labels: transform to get back just string labels from labeled timebins,
  used to evaluate a model
- to_segments: transform to get back segment onsets, offsets, and labels from labeled timebins.
  Inverse of ``from_segments``.
- post-processing transforms that can be used to "clean up" a vector of labeled timebins
  - segment_inds_list_from_class_labels: helper function used to find segments in a vector of labeled timebins
  - remove_short_segments: remove any segment less than a minimum duration
  - take_majority_vote: take a "majority vote" within each segment bounded by the background label,
    and apply the most "popular" label within each segment to all timebins in that segment

Additionally some of the functions have more than one unit test,
where the first tests with simple examples
and the second then tests with real data.
Namely, ``to_labels``, ``to_segments`` and the related functions
``to_labels_with_postprocessing``
and ``to_segments_with_postprocessing``.
Simple examples are used to test expected behavior and edge cases.
Testing with real data complements this.
"""
import copy
import itertools

import numpy as np
import pytest

import vak.common  # for constants
import vak.common.files.spect
import vak.common.labels
import vak.transforms.frame_labels


from ...fixtures.annot import ANNOT_LIST_YARDEN, ANNOT_LIST_NOTMAT, LABELSET_YARDEN, LABELSET_NOTMAT
from ...fixtures.spect import SPECT_LIST_NPZ, SPECT_LIST_MAT


assert len(ANNOT_LIST_YARDEN) == len(SPECT_LIST_MAT), "ANNOT_LIST_YARDEN and SPECT_LIST_MAT are not the same length"

SPECT_LIST_NPZ = copy.deepcopy(SPECT_LIST_NPZ)  # to not mutate the one used by fixtures
ANNOT_LIST_NOTMAT = copy.deepcopy(ANNOT_LIST_NOTMAT)  # to not mutate the one used by fixtures
# make sure ANNOT_LIST_NOTMAT can pair with SPECT_LIST_NPZ
audio_paths_from_spect_list = [
    spect_path.name.replace('.spect.npz', '') for spect_path in SPECT_LIST_NPZ
]
ANNOT_LIST_NOTMAT = [
    annot for annot in ANNOT_LIST_NOTMAT
    if annot.notated_path.name in audio_paths_from_spect_list
]


# define here because we re-use to parametrize multiple tests
# and because we import in .test_transforms
FROM_SEGMENTS_PARAMETRIZE_ARGVALS = list(zip(
    sorted(ANNOT_LIST_YARDEN, key=lambda annot: annot.notated_path.name),
    sorted(SPECT_LIST_MAT, key=lambda spect_path: spect_path.name),
    itertools.repeat(LABELSET_YARDEN)
)) + list(zip(
    sorted(ANNOT_LIST_NOTMAT, key=lambda annot: annot.notated_path.name),
    sorted(SPECT_LIST_NPZ, key=lambda spect_path: spect_path.name),
    itertools.repeat(LABELSET_NOTMAT)
))


@pytest.mark.parametrize(
    'annot, spect_path, labelset',
    FROM_SEGMENTS_PARAMETRIZE_ARGVALS,
)
def test_from_segments(annot, spect_path, labelset):
    labelset = vak.common.converters.labelset_to_set(labelset)
    labelmap = vak.common.labels.to_map(labelset, True)

    spect_dict = vak.common.files.spect.load(spect_path)
    timebins = spect_dict['t']

    try:
        lbls_int = [labelmap[lbl] for lbl in annot.seq.labels]
    except KeyError:
        pytest.skip(
            'Annotation with label not in labelset, would not include in dataset'
        )

    lbl_tb = vak.transforms.frame_labels.from_segments(
        lbls_int,
        annot.seq.onsets_s,
        annot.seq.offsets_s,
        timebins,
        background_label=labelmap[vak.common.constants.DEFAULT_BACKGROUND_LABEL],
    )
    assert lbl_tb.shape == timebins.shape
    assert all(
        [lbl in lbls_int for lbl in np.unique(lbls_int)]
    )


@pytest.mark.parametrize(
    "lbl_tb, labelmap, labels_expected_int",
    [
        (np.array([0, 0, 1, 1, 0, 0, 2, 2, 0, 0]), {vak.common.constants.DEFAULT_BACKGROUND_LABEL: 0, 'a': 1, 'b': 2}, [1, 2]),
        (np.array([0, 0, 1, 1, 0, 0, 2, 2, 0, 0]), {vak.common.constants.DEFAULT_BACKGROUND_LABEL: 0, '1': 1, '2': 2}, [1, 2]),
        (np.array([0, 0, 21, 21, 0, 0, 22, 22, 0, 0]), {vak.common.constants.DEFAULT_BACKGROUND_LABEL: 0, '21': 21, '22': 22}, [21, 22]),
        (np.array([0, 0, 11, 11, 0, 0, 12, 12, 0, 0]), {vak.common.constants.DEFAULT_BACKGROUND_LABEL: 0, '11': 11, '12': 12}, [11, 12]),
    ]
)
def test_to_labels(lbl_tb, labelmap, labels_expected_int):
    # next line, convert all labels to single characters
    # we can easily compare strings we get back with expected;
    # this is what core.eval does
    labelmap = vak.common.labels.multi_char_labels_to_single_char(
        labelmap, skip=(vak.common.constants.DEFAULT_BACKGROUND_LABEL,)
    )
    labelmap_inv = {v: k for k, v in labelmap.items()}
    labels_expected = ''.join([labelmap_inv[lbl_int] for lbl_int in labels_expected_int])

    labels = vak.transforms.frame_labels.to_labels(lbl_tb, labelmap)
    assert labels == labels_expected


# These fail because silent gap durations are too close to the timebin size of the spectrogram
# to be able to recover the original segments by segmenting the labeled timebins.
# ALl these have similar issues, where we can't successfully round trip from labeled timebins to segments,
# because the timebin duration is pretty big (2.7 ms) and there are silent gap durations very close to that
# (e.g. 3 ms), so segments get combined or lost due to rounding error when we do np.min/max below
XFAIL_SPECT_FILES = [
    "llb3_0016_2018_04_23_15_18_14.wav.mat",
    "llb3_0020_2018_04_23_15_30_06.wav.mat",
    "llb3_0053_2018_04_23_17_20_04.wav.mat",
    "llb3_0054_2018_04_23_17_21_23.wav.mat",
    "llb3_0068_2018_04_23_17_34_33.wav.mat",
    "llb3_0071_2018_04_23_17_38_30.wav.mat",  # has zero duration between syllable segments, onsets 54 and 55
    "llb3_0075_2018_04_23_17_41_08.wav.mat",
    "llb3_0074_2018_04_23_17_41_08.wav.mat",

]


@pytest.mark.parametrize(
    'annot, spect_path, labelset',
    FROM_SEGMENTS_PARAMETRIZE_ARGVALS,
)
def test_to_labels_real_data(
        annot, spect_path, labelset
):
    """test that ``to_labels_with_postprocessing`` recovers labels from real data"""
    labelset = vak.common.converters.labelset_to_set(labelset)
    labelmap = vak.common.labels.to_map(labelset)
    # next line, convert all labels to single characters
    # we can easily compare strings we get back with expected;
    # this is what core.eval does
    labelmap = vak.common.labels.multi_char_labels_to_single_char(
        labelmap, skip=(vak.common.constants.DEFAULT_BACKGROUND_LABEL,)
    )
    TIMEBINS_KEY = "t"

    if any(
        [str(spect_path).endswith(spect_file_to_skip)
        for spect_file_to_skip in XFAIL_SPECT_FILES]
    ):
        pytest.xfail(
            "Can't round trip segments -> lbl_tb -> segments "
            "because of small silent gap durations + large time bin durations"
        )

    try:
        lbls_int = [labelmap[lbl] for lbl in annot.seq.labels]
    except KeyError:
        pytest.skip(
            'Annotation with label not in labelset, would not include in dataset'
        )

    timebins = vak.common.files.spect.load(spect_path)[TIMEBINS_KEY]

    lbl_tb = vak.transforms.frame_labels.from_segments(
        lbls_int,
        annot.seq.onsets_s,
        annot.seq.offsets_s,
        timebins,
        background_label=labelmap[vak.common.constants.DEFAULT_BACKGROUND_LABEL],
    )

    labels = vak.transforms.frame_labels.to_labels(
        lbl_tb,
        labelmap,
    )

    labelmap_multi_inv = {v: k for k, v in
                          labelmap.items()}
    labels_expected = "".join(
        [labelmap_multi_inv[lbl_int] for lbl_int in lbls_int]
    )
    assert labels == labels_expected


MAX_ABS_DIFF = 0.003  # milliseconds


@pytest.mark.parametrize(
    'annot, spect_path, labelset',
    FROM_SEGMENTS_PARAMETRIZE_ARGVALS,
)
def test_to_segments_real_data(
        annot, spect_path, labelset
):
    """test that ``to_segments`` recovers onsets, offsets, and labels from real data"""
    labelset = vak.common.converters.labelset_to_set(labelset)
    labelmap = vak.common.labels.to_map(labelset)

    TIMEBINS_KEY = "t"

    if any([str(spect_path).endswith(spect_file_to_skip) for spect_file_to_skip in XFAIL_SPECT_FILES]):
        pytest.xfail(
            "Can't round trip segments -> lbl_tb -> segments "
            "because of small silent gap durations + large time bin durations"
        )

    try:
        lbls_int = [labelmap[lbl] for lbl in annot.seq.labels]
    except KeyError:
        pytest.skip(
            'Annotation with label not in labelset, would not include in dataset'
        )

    timebins = vak.common.files.spect.load(spect_path)[TIMEBINS_KEY]

    lbl_tb = vak.transforms.frame_labels.from_segments(
        lbls_int,
        annot.seq.onsets_s,
        annot.seq.offsets_s,
        timebins,
        background_label=labelmap[vak.common.constants.DEFAULT_BACKGROUND_LABEL],
    )

    labels, onsets_s, offsets_s = vak.transforms.frame_labels.to_segments(
        lbl_tb, labelmap, timebins
    )

    assert np.all(np.char.equal(labels, annot.seq.labels))

    # writing the logic of the function here to test wouldn't make sense
    # but to still test on real data, we can test whether onset_inds
    # is the same length as expected_labels. This should be True
    assert np.all(np.abs(annot.seq.onsets_s - onsets_s) < MAX_ABS_DIFF)
    assert np.all(np.abs(annot.seq.offsets_s - offsets_s) < MAX_ABS_DIFF)


@pytest.mark.parametrize(
        "boundary_labels, boundary_inds_expected",
        [
            (
                np.array([1,0,0,0,1,0,0]),
                np.array([0,4])
            ),
        ]
)
def test_boundary_inds_from_boundary_labels(boundary_labels, boundary_inds_expected):
    boundary_inds = vak.transforms.frame_labels.boundary_inds_from_boundary_labels(
        boundary_labels
    )
    assert np.array_equal(boundary_inds, boundary_inds_expected)


@pytest.mark.parametrize(
        "boundary_labels, expected_exception",
        [
            # 3-d array should raise a ValueError, needs to be row or 1-d
            (
                np.array([[[1,0,0,0,1,0,0]]]),
                ValueError
            ),
            # column vector should raise a ValueError, needs to be row or 1-d
            (
                np.array([[1],[0],[0]]),
                ValueError
            )

        ]
)
def test_boundary_inds_from_boundary_labels(boundary_labels, expected_exception):
    with pytest.raises(expected_exception):
        vak.transforms.frame_labels.functional.segment_inds_list_from_boundary_labels(
            boundary_labels
        )


@pytest.mark.parametrize(
    "frame_labels, seg_inds_list_expected",
    [
        (np.asarray([0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0]), [np.array([4, 5, 6, 7])]),
        # assert works when segment is at start of lbl_tb
        (np.asarray([1, 1, 1, 1, 0, 0, 0, 0]), [np.array([0, 1, 2, 3])]),
        # assert works with multiple segments
        (
            np.array([0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 2, 1, 0, 0]),
            [np.array([3, 4, 5]), np.array([9, 10, 11])],
        ),
        # assert works when a segment is at end of lbl_tb
        (
            np.array([0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 2, 1]),
            [np.array([3, 4, 5]), np.array([9, 10, 11])],
        ),
    ],
)
def test_segment_inds_list_from_class_labels(frame_labels, seg_inds_list_expected):
    """Test ``to_inds`` works as expected"""
    UNLABELED = 0

    seg_inds_list = vak.transforms.frame_labels.segment_inds_list_from_class_labels(
        frame_labels=frame_labels, background_label=UNLABELED
    )
    assert np.array_equal(seg_inds_list, seg_inds_list_expected)


@pytest.mark.parametrize(
    'lbl_tb, unlabeled, timebin_dur, min_segment_dur, lbl_tb_expected',
    [
        # should remove the 1 at the end if lbl_tb since it's a segment with dur < 0.002
        (
            np.asarray([0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 1, 0]),
            0,
            0.001,
            0.002,
            np.asarray([0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0])
        ),
        # should **not** remove a segment with dur == 0.002
        (
            np.asarray([0, 0, 0, 0, 1, 1, 1, 1, 0, 1, 1, 0]),
            0,
            0.001,
            0.002,
            np.asarray([0, 0, 0, 0, 1, 1, 1, 1, 0, 1, 1, 0])
        )
    ]
)
def test_remove_short_segments(lbl_tb, unlabeled, timebin_dur, min_segment_dur, lbl_tb_expected):
    """Test ``remove_short_segments`` works as expected"""
    segment_inds_list = vak.transforms.frame_labels.segment_inds_list_from_class_labels(
        lbl_tb, background_label=unlabeled
    )
    lbl_tb_tfm, segment_inds_list_out = vak.transforms.frame_labels.remove_short_segments(
        lbl_tb,
        segment_inds_list,
        timebin_dur=timebin_dur,
        min_segment_dur=min_segment_dur,
        background_label=unlabeled,
    )
    assert np.array_equal(lbl_tb_tfm, lbl_tb_expected)


@pytest.mark.parametrize(
    "lbl_tb_in, unlabeled, lbl_tb_expected",
    [
        # should do nothing when a labeled segment has all the same labels
        (
            np.asarray([0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0]),
            0,
            np.asarray([0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0]),
        ),
        (
            np.array([0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 2, 1, 0, 0]),
            0,
            np.array([0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 0]),
        ),
        # test MajorityVote works when there is no vak.common.constants.DEFAULT_BACKGROUND_LABEL segment at start of vector
        (
            np.asarray([1, 1, 2, 1, 0, 0, 0, 0]),
            0,
            np.asarray([1, 1, 1, 1, 0, 0, 0, 0])
        ),
        # test MajorityVote works when there is no vak.common.constants.DEFAULT_BACKGROUND_LABEL segment at end of vector
        (
            np.array([0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 2, 1]),
            0,
            np.array([0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 1, 1]),
        ),
        # test that a tie results in lowest value class winning, default behavior of scipy.stats.mode
        (
            np.array([0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 1, 2, 2]),
            0,
            np.array([0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1]),
        ),
    ],
)
def test_majority_vote(lbl_tb_in, unlabeled, lbl_tb_expected):
    """Test ``majority_vote`` works as expected"""
    segment_inds_list = vak.transforms.frame_labels.segment_inds_list_from_class_labels(
        lbl_tb_in, background_label=unlabeled
    )
    lbl_tb_maj_vote = vak.transforms.frame_labels.take_majority_vote(
        lbl_tb_in, segment_inds_list
    )
    assert np.array_equal(lbl_tb_maj_vote, lbl_tb_expected)


# ---- define these constants here we use with pytest.mark.parametrize
# so that we can import them in .test_transforms as well
TIMEBIN_DUR_FOR_PARAMETRIZE = 0.001
UNLABELED_LABEL = 0
POSTPROCESS_PARAMS_ARGVALS = [
    # test case where we apply *neither* of the transforms
    (
            np.asarray([0, 1, 1, 0, 2, 2, 0, 3, 3, 0, 0, 4, 4, 0, 0]),
            None,
            False,
            np.asarray([0, 1, 1, 0, 2, 2, 0, 3, 3, 0, 0, 4, 4, 0, 0]),
    ),
    # test case where we apply *neither* of the transforms, and one segment is at end of lbl_tb
    (
            np.asarray([0, 1, 1, 0, 2, 2, 0, 3, 3, 0, 0, 4, 4, 4, 4]),
            None,
            False,
            np.asarray([0, 1, 1, 0, 2, 2, 0, 3, 3, 0, 0, 4, 4, 4, 4]),
    ),
    # ---- start of test cases for majority vote
    # test MajorityVote does nothing when a labeled segment has all the same labels
    (
        np.asarray([0, 1, 1, 0, 2, 2, 2, 2, 0, 0, 0, 0]),
        None,
        True,
        np.asarray([0, 1, 1, 0, 2, 2, 2, 2, 0, 0, 0, 0]),
    ),
    # test majority vote
    (
        np.array([0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 2, 1, 0, 0]),
        None,
        True,
        # majority vote converts second segment to label "a"
        np.array([0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 1, 1, 0, 0]),
    ),
    # test MajorityVote works when there is no vak.common.constants.DEFAULT_BACKGROUND_LABEL segment at start of vector
    (
        np.array([1, 1, 2, 1, 0, 0, 0, 0]),
        None,
        True,
        np.array([1, 1, 1, 1, 0, 0, 0, 0]),
    ),
    # test MajorityVote works when there is no vak.common.constants.DEFAULT_BACKGROUND_LABEL segment at end of vector
    (
        np.array([0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 2, 1]),
        None,
        True,
        np.array([0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 1, 1]),
    ),
    # test that a tie results in lowest value class winning, default behavior of scipy.stats.mode
    (
        np.array([0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 1, 2, 2]),
        None,
        True,
        np.array([0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1]),
    ),
    # test that majority vote just returns lbl_tb untouched when everything is unlabeled
    (
        np.ones(4000).astype(int) * UNLABELED_LABEL,  # i.e. all zeros, but being explicit here
        None,
        True,
        np.ones(4000).astype(int) * UNLABELED_LABEL,
    ),
    # ---- start of test cases for min segment dur
    # should remove a segment with dur < min_segment_dur
    (
        np.asarray([0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 1, 0]),
        0.002,
        False,
        np.asarray([0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0]),
    ),
    # should **not** remove a segment with dur == 0.002
    (
        np.asarray([0, 0, 0, 0, 1, 1, 1, 1, 0, 1, 1, 0]),
        0.002,
        False,
        np.asarray([0, 0, 0, 0, 1, 1, 1, 1, 0, 1, 1, 0]),
    ),
    # test min_segment_dur returns all Nones when all segments are less than min segment dur
    (
        np.array([0, 0, 0, 0, 0, 0, 1, 1, 2,
                  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                  1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, ]),
        0.025,  # notice segment dur, 25ms. Realistic value but will remove all segments in lbl_tb
        False,
        np.ones(36).astype(int) * UNLABELED_LABEL,  # i.e. all zeros, but being explicit here
    ),
]

# now rewrite but with args in order for function call:
POSTPROCESS_PARAMS_ARGVALS = [
    argvals[:1] + (TIMEBIN_DUR_FOR_PARAMETRIZE, UNLABELED_LABEL) + argvals[1:]
    for argvals in POSTPROCESS_PARAMS_ARGVALS
]


@pytest.mark.parametrize(
    'lbl_tb, timebin_dur, background_label, min_segment_dur, majority_vote, lbl_tb_expected',
    POSTPROCESS_PARAMS_ARGVALS
)
def test_postprocess(lbl_tb, timebin_dur, background_label, min_segment_dur, majority_vote, lbl_tb_expected):
    """Test that ``trasnforms.frame_labels.postprocess`` works as expected.
    Specifically test that we recover an expected string of labels,
    as would be used to compute edit distance."""
    lbl_tb = vak.transforms.frame_labels.postprocess(
        lbl_tb,
        timebin_dur=timebin_dur,
        background_label=UNLABELED_LABEL,
        majority_vote=majority_vote,
        min_segment_dur=min_segment_dur,
    )

    assert np.all(np.equal(lbl_tb, lbl_tb_expected))



from . import curvefit, dirname, frame_classification, learncurve
from .learncurve import learning_curve

__all__ = [
    "curvefit",
    "dirname",
    "frame_classification",
    "learncurve",
    "learning_curve",
]



"""tests for vak.config.spect_params module"""
import pytest

import vak.config.spect_params

from .attr_helpers import simple_attr


# approaching to testing validators adopted from
# https://github.com/python-attrs/attrs/blob/a025629e36440dcc27aee0ee5b04d6523bcc9931/tests/test_validators.py#L338
def test_freq_cutoffs_not_two_values_raises():
    """test that a ValueError is raised when freq_cutoffs is not just two values"""
    freq_cutoff_attrib = simple_attr("freq_cutoffs")

    with pytest.raises(ValueError):
        vak.config.spect_params.freq_cutoffs_validator(
            instance=None, attribute=freq_cutoff_attrib, value=[0]
        )

    with pytest.raises(ValueError):
        vak.config.spect_params.freq_cutoffs_validator(
            instance=None, attribute=freq_cutoff_attrib, value=[0, 10, 100]
        )


def test_invalid_transform_type_raises():
    """test that an invalid spectrogram transform type raises a ValueError"""
    transform_type_attrib = simple_attr("transform_type")

    with pytest.raises(ValueError):
        vak.config.spect_params.is_valid_transform_type(
            instance=None,
            attribute=transform_type_attrib,
            value="log",  # not in vak.config.spect_params.VALID_TRANSFORM_TYPES
        )


def test_freq_cutoffs_wrong_order_raises():
    """test that a ValueError is raised when freq_cutoffs[0] > freq_cutoffs[1]"""
    freq_cutoff_attrib = simple_attr("freq_cutoffs")

    with pytest.raises(ValueError):
        vak.config.spect_params.freq_cutoffs_validator(
            instance=None, attribute=freq_cutoff_attrib, value=[10000, 500]
        )


class TestSpectParamsConfig:
    @pytest.mark.parametrize(
        'config_dict',
        [
            {'fft_size': 512, 'step_size': 64, 'freq_cutoffs': [500, 10000], 'thresh': 6.25, 'transform_type': 'log_spect'},
        ]
    )
    def test_init(self, config_dict):
        spect_params_config = vak.config.SpectParamsConfig(**config_dict)
        assert isinstance(spect_params_config, vak.config.spect_params.SpectParamsConfig)

    def test_with_real_config(self, a_generated_config_dict):
        if "spect_params" in a_generated_config_dict['prep']:
            spect_config_dict = a_generated_config_dict['prep']['spect_params']
        else:
            pytest.skip("No spect params in config")
        spect_params_config = vak.config.spect_params.SpectParamsConfig(**spect_config_dict)
        assert isinstance(spect_params_config, vak.config.spect_params.SpectParamsConfig)



import pandas as pd
import pytest

import vak


@pytest.mark.parametrize(
    'config_type, model_name, audio_format, spect_format, annot_format, input_type',
    [
        ('train', 'TweetyNet', 'cbin', None, 'notmat', 'spect'),
        ('predict', 'TweetyNet', 'cbin', None, 'notmat', 'spect'),
        ('eval', 'TweetyNet', 'cbin', None, 'notmat', 'spect'),
        ('train', 'TweetyNet', None, 'mat', 'yarden', 'spect'),
        ('learncurve', 'TweetyNet', 'cbin', None, 'notmat', 'spect'),
        # TODO: add audio cases
    ]
)
def test_assign_samples_to_splits(
        config_type, model_name, audio_format, spect_format, annot_format,
        input_type, tmp_path, specific_config_toml_path, specific_source_files_df,
):
    toml_path = specific_config_toml_path(
        config_type,
        model_name,
        annot_format,
        audio_format,
        spect_format,
    )

    cfg = vak.config.Config.from_toml_path(toml_path)

    # ---- set up ----
    tmp_dataset_path = tmp_path / 'dataset_dir'
    tmp_dataset_path.mkdir()

    purpose = config_type

    source_files_df = specific_source_files_df(
        config_type,
        model_name,
        annot_format,
        audio_format,
        spect_format,
    )

    out = vak.prep.frame_classification.assign_samples_to_splits(
        purpose,
        source_files_df,
        tmp_dataset_path,
        cfg.prep.train_dur,
        cfg.prep.val_dur,
        cfg.prep.test_dur,
        cfg.prep.labelset,
    )

    assert isinstance(out, pd.DataFrame)
    assert 'split' in out.columns
    if purpose == 'predict':
        assert all(val == 'predict' for val in out['split'].values)
    elif purpose == 'eval':
        assert all(val == 'test' for val in out['split'].values)
    else:
        split_vals = out['split'].values.tolist()
        assert all(
            [
                split_name in split_vals
                for split_name in ('train', 'val', 'test')
                if hasattr(cfg.prep, f'{split_name}_dur') and getattr(cfg.prep, f'{split_name}_dur') is not None
            ]
        )



"""Helper functions used with frame classification datasets."""

from __future__ import annotations

from ... import common
from . import constants


def sample_ids_array_filename_for_subset(subset: str) -> str:
    """Returns name of sample IDs array file for a subset of the training data."""
    return constants.SAMPLE_IDS_ARRAY_FILENAME.replace(
        ".npy", f"-{subset}.npy"
    )


def inds_in_sample_array_filename_for_subset(subset: str) -> str:
    """Returns name of inds in sample array file for a subset of the training data."""
    return constants.INDS_IN_SAMPLE_ARRAY_FILENAME.replace(
        ".npy", f"-{subset}.npy"
    )


def load_frames(frames_path, input_type):
    """Helper function that loads "frames",
    the input to the frame classification model.
    Loads audio or spectrogram, depending on
    :attr:`self.input_type`.
    This function assumes that audio is in wav format
    and spectrograms are in npz files.
    """
    if input_type == "audio":
        frames, _ = common.constants.AUDIO_FORMAT_FUNC_MAP[
            constants.FRAME_CLASSIFICATION_DATASET_AUDIO_FORMAT
        ](frames_path)
    elif input_type == "spect":
        spect_dict = common.files.spect.load(frames_path)
        frames = spect_dict[common.constants.SPECT_KEY]
    return frames



"""Helper functions for moving and modifying configs"""
import logging
import pathlib
import shutil

import tomlkit

import vak.cli.prep
from . import constants


logger = logging.getLogger(__name__)


def copy_config_files():
    """copy config files from setup to data_for_tests/configs

    the copied files are the ones that get modified when this setup script runs,
    while the originals in this directory remain unchanged.
    """
    logger.info(
        f"Making directory to copy config files:\n{constants.GENERATED_TEST_CONFIGS_ROOT}"
    )
    constants.GENERATED_TEST_CONFIGS_ROOT.mkdir(parents=True)

    logger.info(
        "Copying config files run to generate test data from ./tests/data_for_tests/configs to "
        f"{constants.GENERATED_TEST_CONFIGS_ROOT}"
    )


    copied_configs = []

    for config_metadata in constants.CONFIG_METADATA:
        toml_path = constants.TEST_CONFIGS_ROOT / config_metadata.filename
        if not toml_path.exists():
            raise FileNotFoundError(f"{toml_path} not found")

        dst = constants.GENERATED_TEST_CONFIGS_ROOT.joinpath(toml_path.name)
        logger.info(f"\tCopying '{toml_path.name}'")
        shutil.copy(src=toml_path, dst=dst)
        copied_configs.append(dst)

    return copied_configs


def add_dataset_path_from_prepped_configs():
    """This helper function goes through all configs in
    :data:`vaktestdata.constants.CONFIG_METADATA`
    and for any that have a filename for the attribute
    "use_dataset_from_config", it sets the key 'path' in the 'dataset' table
    in the config file that the metadata corresponds to
    to the same value from the file specified by the attribute.
    """
    configs_to_change = [
        config_metadata
        for config_metadata in constants.CONFIG_METADATA
        if config_metadata.use_dataset_from_config is not None
    ]

    for config_metadata in configs_to_change:
        config_to_change_path = constants.GENERATED_TEST_CONFIGS_ROOT / config_metadata.filename
        if config_metadata.config_type == 'train_continue':
            table_to_add_dataset = 'train'
        else:
            table_to_add_dataset = config_metadata.config_type

        config_dataset_path = constants.GENERATED_TEST_CONFIGS_ROOT / config_metadata.use_dataset_from_config

        config_dict = vak.config.load._load_toml_from_path(config_dataset_path)
        # next line, we can't use `section` here because we could get a KeyError,
        # e.g., when the config we are rewriting is an ``[vak.eval]`` config, but
        # the config we are getting the dataset from is a ``[vak.train]`` config.
        # so instead we use `purpose_from_toml` to get the `purpose`
        # of the config we are getting the dataset from.
        dataset_config_section = vak.cli.prep.purpose_from_toml(config_dict)
        dataset_path = config_dict[dataset_config_section]['dataset']['path']

        # we open config using tomlkit so we can add path to dataset table in style-preserving way
        with config_to_change_path.open('r') as fp:
            tomldoc = tomlkit.load(fp)
        if 'dataset' not in tomldoc['vak'][table_to_add_dataset]:
            dataset_table = tomlkit.table()
            tomldoc["vak"][table_to_add_dataset].add("dataset", dataset_table)
        tomldoc["vak"][table_to_add_dataset]["dataset"].add("path", str(dataset_path))
        with config_to_change_path.open("w") as fp:
            tomlkit.dump(tomldoc, fp)


def fix_options_in_configs(config_metadata_list, command, single_train_result=True):
    """Fix values assigned to options in predict and eval configs.

    Need to do this because both predict and eval configs have options
    that can only be assigned *after* running the corresponding `train` config.
    """
    if command not in ('eval', 'predict', 'train_continue'):
        raise ValueError(
            f'invalid command to fix config options: {command}'
        )

    for config_metadata in config_metadata_list:
        config_to_fix = constants.GENERATED_TEST_CONFIGS_ROOT / config_metadata.filename
        config_to_use_result_from = constants.GENERATED_TEST_CONFIGS_ROOT / config_metadata.use_result_from_config

        # now use the config to find the results dir and get the values for the options we need to set
        # which are checkpoint_path, frames_standardizer_path, and labelmap_path
        with config_to_use_result_from.open("r") as fp:
            config_toml = tomlkit.load(fp)
        root_results_dir = pathlib.Path(config_toml["vak"]["train"]["root_results_dir"])
        results_dir = sorted(root_results_dir.glob("results_*"))
        if len(results_dir) > 1:
            if single_train_result:
                raise ValueError(
                    f"Did not find just a single results directory in root_results_dir from train_config:\n"
                    f"{config_to_use_result_from}"
                    f"root_results_dir was: {root_results_dir}"
                    f'Matches for "results_*" were: {results_dir}'
                )
            else:
                results_dir = results_dir[-1]
        elif len(results_dir) == 1:
            results_dir = results_dir[0]
        else:
            raise ValueError(
                f"Did not find a results directory in root_results_dir from train_config:\n"
                f"{config_to_use_result_from}"
                f"root_results_dir was:\n{root_results_dir}"
                f'Matches for "results_*" were:\n{results_dir}'
            )

        # these are the only options whose values we need to change
        # and they are the same for both predict and eval
        checkpoint_path = sorted(results_dir.glob("**/checkpoints/checkpoint.pt"))[0]
        if 'standardize_frames' in config_toml["vak"]['train'] and config_toml["vak"]['train']['standardize_frames']:
            frames_standardizer_path = sorted(results_dir.glob("FramesStandardizer"))[0]
        else:
            frames_standardizer_path = None

        labelmap_path = sorted(results_dir.glob("labelmap.json"))
        if len(labelmap_path) == 1:
            labelmap_path = labelmap_path[0]
        elif len(labelmap_path) == 0:
            labelmap_path = None
        else:
            raise ValueError(
                "Invalid number of labelmap.json files from results_dir for train config:\n"
                f"{config_to_use_result_from}.\n"
                f"Results dir: {results_dir}"
                f"labelmap_path(s) found by globbing: {labelmap_path}"
            )

        # now add these values to corresponding options in predict / eval config
        with config_to_fix.open("r") as fp:
            config_toml = tomlkit.load(fp)

        if command == 'train_continue':
            table = 'train'
        else:
            table = command

        config_toml["vak"][table]["checkpoint_path"] = str(checkpoint_path)
        if frames_standardizer_path:
            config_toml["vak"][table]["frames_standardizer_path"] = str(frames_standardizer_path)
        else:
            if 'frames_standardizer_path' in config_toml["vak"][table]:
                # remove any existing 'frames_standardizer_path' option
                del config_toml["vak"][table]["frames_standardizer_path"]
        if command != 'train_continue':  # train always gets labelmap from dataset dir, not from a config option
            if labelmap_path is not None:
                config_toml["vak"][table]["labelmap_path"] = str(labelmap_path)

        with config_to_fix.open("w") as fp:
            tomlkit.dump(config_toml, fp)



import inspect

import pytest

import vak.models.registry
from .conftest import (
    MockModel,
    MockEncoderDecoderModel,
    MockModelFamily,
    UnregisteredMockModelFamily,
)
from .test_definition import (
    TweetyNetDefinition,
)


def test_model_family():
    """Test that :func:`vak.models.registry.model_family`
    adds a model family to the registry"""
    # we make this copy so that we don't register UnregisteredMockModelFamily;
    # we need that class to stay unregistered for other tests
    ModelFamilyCopy = type('ModelFamilyCopy',
                           UnregisteredMockModelFamily.__bases__,
                           dict(UnregisteredMockModelFamily.__dict__))
    assert ModelFamilyCopy.__name__ not in vak.models.registry.MODEL_FAMILY_REGISTRY
    vak.models.registry.model_family(ModelFamilyCopy)
    assert ModelFamilyCopy.__name__ in vak.models.registry.MODEL_FAMILY_REGISTRY
    assert vak.models.registry.MODEL_FAMILY_REGISTRY[ModelFamilyCopy.__name__] == ModelFamilyCopy


@pytest.mark.parametrize(
    'family, definition',
    [
        (MockModelFamily, MockModel),
        (MockModelFamily, MockEncoderDecoderModel),
    ]
)
def test_register_model(family, definition):
    """Test that :func:`vak.models.registry.register_model`
    adds a model to the registry"""
    # to set up, we repeat what :func:`vak.models.decorator.model` does
    model_name = definition.__name__
    model_factory = vak.models.factory.ModelFactory(
        definition,
        family
    )
    model_factory.__name__ = model_name

    assert model_name not in vak.models.registry.MODEL_REGISTRY
    vak.models.registry.register_model(model_factory)
    assert model_name in vak.models.registry.MODEL_REGISTRY
    assert vak.models.registry.MODEL_REGISTRY[model_name] == model_factory
    del vak.models.registry.MODEL_REGISTRY[model_name]  # so this test doesn't fail for the second case


def test_register_model_raises_family():
    """Test that :func:`vak.models.registry.register_model`
    raises an error if parent class is not in model_family_classes"""
    # to set up, we repeat what :func:`vak.models.decorator.model` does,
    # but notice that we use an unregistered model family
    model_factory = vak.models.ModelFactory(
        MockModel,
        UnregisteredMockModelFamily,
    )

    with pytest.raises(TypeError):
        vak.models.registry.register_model(model_factory)


@pytest.mark.parametrize(
    'family, definition',
    [
        (vak.models.FrameClassificationModel, TweetyNetDefinition),
    ]
)
def test_register_model_raises_registered(family, definition):
    """Test that :func:`vak.models.registry.register_model`
    raises an error if a class is already registered"""
    # to set up, we repeat what :func:`vak.models.decorator.model` does
    model_factory = vak.models.ModelFactory(
        definition,
        family
    )

    # NOTE we replace 'Definition' with an empty string
    # so that the name clashes with an existing model name
    model_factory.__name__ = definition.__name__.replace('Definition', '')

    with pytest.raises(ValueError):
        vak.models.registry.register_model(model_factory)


def test___get_attr__MODEL_FAMILY_FROM_NAME():
    assert hasattr(vak.models.registry, 'MODEL_FAMILY_FROM_NAME')

    model_family_from_name_dict = getattr(vak.models.registry, 'MODEL_FAMILY_FROM_NAME')
    assert isinstance(model_family_from_name_dict, dict)

    for model_name, model_factory in vak.models.registry.MODEL_REGISTRY.items():
        model_family = model_factory.family
        family_name = model_family.__name__
        assert model_family_from_name_dict[model_name] == family_name


def test___get_attr__MODEL_NAMES():
    assert hasattr(vak.models.registry, 'MODEL_NAMES')
    model_names_list = getattr(vak.models.registry, 'MODEL_NAMES')
    assert isinstance(model_names_list, list)
    for model_name in vak.models.registry.MODEL_REGISTRY.keys():
            assert model_name in model_names_list



from . import fixtures
# keep this import here, we need it for fixtures
from .fixtures import *


def by_slow_marker(item):
    return 1 if item.get_closest_marker('slow') is None else 0



def pytest_addoption(parser):
    parser.addoption('--dtype', action="store", default="float32")
    parser.addoption('--slow-last', action='store_true', default=False)


def pytest_collection_modifyitems(items, config):
    if config.getoption('--slow-last'):
        items.sort(key=by_slow_marker, reverse=True)


def pytest_generate_tests(metafunc):
    if 'dtype_name' in metafunc.fixturenames:
        raw_value = metafunc.config.getoption('--dtype')
        if raw_value == 'all':
            dtype_names = list(fixtures.torch.TEST_DTYPES.keys())
        else:
            dtype_names = raw_value.split(',')
        if dtype_names is not None:
            metafunc.parametrize('dtype_name', dtype_names)



"""Modules that perform neural network convolutions."""

import torch
from torch.nn import functional as F


# NOTE: added 2023-03-04
# in this class, we detect when one extra pixel should be added on the bottom or right
# and specifically pad those, see line 75, ``if rows_odd or cols_odd:``.
class Conv2dTF(torch.nn.Conv2d):
    """Conv2d with padding behavior from Tensorflow

    Adapted from
    https://github.com/mlperf/inference/blob/16a5661eea8f0545e04c86029362e22113c2ec09/others/edge/object_detection/ssd_mobilenet/pytorch/utils.py#L40
    as referenced in this issue:
    https://github.com/pytorch/pytorch/issues/3867#issuecomment-507025011

    Used to maintain behavior of original implementation of TweetyNet that used Tensorflow 1.0 low-level API.

    Note there are issues with SAME convolution as performed by Tensorflow.
    See https://gist.github.com/Yangqing/47772de7eb3d5dbbff50ffb0d7a98964.
    """

    PADDING_METHODS = ("VALID", "SAME")

    def __init__(self, *args, **kwargs):
        # remove 'padding' from ``kwargs`` to avoid bug in ``torch`` => 1.7.2
        # see https://github.com/yardencsGitHub/tweetynet/issues/166
        kwargs_super = {k: v for k, v in kwargs.items() if k != "padding"}
        super(Conv2dTF, self).__init__(*args, **kwargs_super)
        padding = kwargs.get("padding", "SAME")
        if not isinstance(padding, str):
            raise TypeError(
                f"value for 'padding' argument should be a string, one of: {self.PADDING_METHODS}"
            )
        padding = padding.upper()
        if padding not in self.PADDING_METHODS:
            raise ValueError(
                f"value for 'padding' argument must be one of '{self.PADDING_METHODS}' but was: {padding}"
            )
        self.padding = padding

    def _compute_padding(self, input, dim):
        input_size = input.size(dim + 2)
        filter_size = self.weight.size(dim + 2)
        effective_filter_size = (filter_size - 1) * self.dilation[dim] + 1
        out_size = (input_size + self.stride[dim] - 1) // self.stride[dim]
        total_padding = max(
            0,
            (out_size - 1) * self.stride[dim]
            + effective_filter_size
            - input_size,
        )
        additional_padding = int(total_padding % 2 != 0)

        return additional_padding, total_padding

    def forward(self, input):
        if self.padding == "VALID":
            return F.conv2d(
                input,
                self.weight,
                self.bias,
                self.stride,
                padding=0,
                dilation=self.dilation,
                groups=self.groups,
            )
        elif self.padding == "SAME":
            rows_odd, padding_rows = self._compute_padding(input, dim=0)
            cols_odd, padding_cols = self._compute_padding(input, dim=1)
            if rows_odd or cols_odd:
                input = F.pad(input, [0, cols_odd, 0, rows_odd])

            return F.conv2d(
                input,
                self.weight,
                self.bias,
                self.stride,
                padding=(padding_rows // 2, padding_cols // 2),
                dilation=self.dilation,
                groups=self.groups,
            )



"""TweetyNet model [1]_.

.. [1] TweetyNet was described in:
   Cohen, Y., Nicholson, D. A., Sanchioni, A., Mallaber, E. K., Skidanova, V., & Gardner, T. J. (2022).
   Automated annotation of birdsong with a neural network that segments spectrograms. eLife 11: e63853.
   Paper: https://elifesciences.org/articles/63853
   Code: https://github.com/yardencsGitHub/tweetynet
"""

from __future__ import annotations

import torch

from .. import metrics, nets, nn
from .decorator import model
from .frame_classification_model import FrameClassificationModel


@model(family=FrameClassificationModel)
class TweetyNet:
    """TweetyNet model, as described in
    Cohen, Y., Nicholson, D. A., Sanchioni, A., Mallaber, E. K., Skidanova, V., & Gardner, T. J. (2022).
    Automated annotation of birdsong with a neural network that segments spectrograms. Elife, 11, e63853.
    https://elifesciences.org/articles/63853.

    Code adapted from
    https://github.com/yardencsGitHub/tweetynet.

    Attributes
    ----------
    network : vak.nets.TweetyNet
        Convolutional-bidirectional LSTM neural network architecture.
    loss: torch.nn.CrossEntropyLoss
        Standard cross-entropy loss
    optimizer: torch.optim.Adam
        Adam optimizer.
    metrics: dict
        Mapping string names to the following metrics:
        ``vak.metrics.Accuracy``, ``vak.metrics.Levenshtein``,
        ``vak.metrics.CharacterErrorRate``, ``torch.nn.CrossEntropyLoss``.

    Notes
    -----
    TweetyNet was described in [1]_.

    ``TweetyNet`` is a type of windowed frame classification model,
    and this version built into ``vak`` relies on the
    ``FrameClassificationModel`` class.

    References
    ----------
    .. [1] Cohen, Y., Nicholson, D. A., Sanchioni, A., Mallaber, E. K., Skidanova, V., & Gardner, T. J. (2022).
       Automated annotation of birdsong with a neural network that segments spectrograms. eLife 11: e63853.
       Paper: https://elifesciences.org/articles/63853
       Code: https://github.com/yardencsGitHub/tweetynet
    """

    network = nets.TweetyNet
    loss = nn.loss.CrossEntropyLoss
    optimizer = torch.optim.Adam
    metrics = {
        "acc": metrics.Accuracy,
        "levenshtein": metrics.Levenshtein,
        "character_error_rate": metrics.CharacterErrorRate,
        "loss": nn.loss.CrossEntropyLoss,
    }
    default_config = {"optimizer": {"lr": 0.003}}



"""tests for vak.config.load module"""
import tomlkit

import vak.config.load


def test__tomlkit_to_pop(a_generated_config_path):
    with a_generated_config_path.open('r') as fp:
        tomldoc = tomlkit.load(fp)
    out = vak.config.load._tomlkit_to_popo(tomldoc)
    assert isinstance(out, dict)
    assert list(out.keys()) == ["vak"]


def test__load_from_toml_path(a_generated_config_path):
    config_dict = vak.config.load._load_toml_from_path(a_generated_config_path)

    assert isinstance(config_dict, dict)

    with a_generated_config_path.open('r') as fp:
        tomldoc = tomlkit.load(fp)
    config_dict_raw = vak.config.load._tomlkit_to_popo(tomldoc)

    assert len(list(config_dict.keys())) == len(list(config_dict_raw["vak"].keys()))



"""tests for ``vak.prep.spectrogram_dataset.spect_helper`` module"""
from pathlib import Path

import pandas as pd
import pytest

import vak.prep.spectrogram_dataset.spect_helper
import vak.common.files.spect


def spect_paths_from_df_as_paths(dataset_df):
    return [Path(spect_path) for spect_path in dataset_df["spect_path"]]


def assert_expected_spect_paths_in_dataframe(
    spect_paths_from_df, spect_format, expected_spect_paths, not_expected_spect_paths=None
):
    """Tests that a dataframe ``dataset_df`` contains one file
    for each path in ``expected_spect_paths`` in its ``spect_path`` column,
    and only those paths.

    Parameters
    ----------
    dataset_df : pandas.Dataframe
        created by vak.prep.spectrogram_dataset.spect_helper.make_dataframe_of_spect_files
    expected_spect_paths : list
        of paths to spectrogram files, that **should** be in dataset_df.spect_path column
    not_expected_spect_paths : list
        of paths to spectrogram files, that should **not** be in dataset_df.spect_path column
    """
    spect_file_names_from_df = [spect_path.name for spect_path in spect_paths_from_df]
    expected_spectfile_names = [
        spect_path.name for spect_path in expected_spect_paths
    ]

    assert all(
        [spect_file_name_from_df in expected_spectfile_names
         for spect_file_name_from_df in spect_file_names_from_df]
    )

    # test that **only** expected paths were in DataFrame
    if not_expected_spect_paths is not None:
        not_expected_spectfile_names = [
            spect_path.name for spect_path in not_expected_spect_paths
        ]
        assert all(
            [spect_file_name_from_df not in not_expected_spectfile_names
             for spect_file_name_from_df in spect_file_names_from_df]
        )


@pytest.mark.parametrize(
    "spect_format, annot_format, spect_ext, labelset, arg_to_test",
    [
        ("mat", "yarden", None, True, 'spect_dir'),
        ("npz", "notmat", ".spect.npz", True, 'spect_dir'),
        ("mat", "yarden", None, False, 'spect_dir'),
        ("npz", "notmat", ".spect.npz", False, 'spect_dir'),
        ("mat", None, None, False, 'spect_dir'),
        ("npz", None, ".spect.npz", False, 'spect_dir'),

        ("mat", "yarden", None, True, 'spect_files'),
        ("npz", "notmat", ".spect.npz", True, 'spect_files'),
        ("mat", "yarden", None, False, 'spect_files'),
        ("npz", "notmat", ".spect.npz", False, 'spect_files'),
        ("mat", None, None, False, 'spect_files'),
        ("npz", None, ".spect.npz", False, 'spect_files'),
    ],
)
def test_make_dataframe_of_spect_files(
    spect_format,
    annot_format,
    spect_ext,
    labelset,
    arg_to_test,
    specific_spect_dir,
    specific_spect_list,
    specific_annot_list,
    specific_labelset,
    tmp_path,
):
    """Test that ``vak.prep.spectrogram_dataset.spect_helper.make_dataframe_of_spect_files`` works
    when we point it at directory + give it list of annotations"""
    if arg_to_test == 'spect_dir':
        spect_dir = specific_spect_dir(spect_format)
        spect_files = None
    elif arg_to_test == 'spect_files':
        spect_dir = None
        spect_files = specific_spect_list(spect_format)

    if labelset:
        labelset = specific_labelset(annot_format)
    else:
        labelset = None

    if annot_format:
        annot_list = specific_annot_list(annot_format)
    else:
        annot_list = None

    if spect_format == "mat":
        spect_output_dir = tmp_path
    else:
        spect_output_dir = None

    dataset_df = vak.prep.spectrogram_dataset.spect_helper.make_dataframe_of_spect_files(
        spect_format=spect_format,
        spect_dir=spect_dir,
        spect_files=spect_files,
        labelset=labelset,
        annot_list=annot_list,
        annot_format=annot_format,
        spect_ext=spect_ext,
    )
    assert type(dataset_df) == pd.DataFrame

    spect_paths_from_df = spect_paths_from_df_as_paths(dataset_df)
    if labelset:
        expected_spect_list = specific_spect_list(
            spect_format, "all_labels_in_labelset"
        )
        not_expected_spect_list = specific_spect_list(
            spect_format, "labels_not_in_labelset"
        )
    else:
        expected_spect_list = specific_spect_list(spect_format)
        not_expected_spect_list = None

    assert_expected_spect_paths_in_dataframe(
        spect_paths_from_df, spect_format,
        expected_spect_list, not_expected_spect_list
    )

    if spect_dir is not None:
        assert all(
            [spect_path.parent == spect_dir for spect_path in spect_paths_from_df]
        )


def test_make_dataframe_of_spect_files_no_spect_dir_files_or_map_raises(annot_list_yarden):
    """test that calling ``make_dataframe_of_spect_files`` without one of:
    spect dir, spect files, or spect files/annotations mapping
    raises ValueError"""
    with pytest.raises(ValueError):
        vak.prep.spectrogram_dataset.spect_helper.make_dataframe_of_spect_files(
            spect_format="mat",
            spect_dir=None,
            spect_files=None,
            annot_list=annot_list_yarden,
            annot_format="yarden",
        )


def test_make_dataframe_of_spect_files_invalid_spect_format_raises(spect_dir_mat, annot_list_yarden):
    """test that calling ``make_dataframe_of_spect_files`` with an invalid spect format raises a ValueError"""
    with pytest.raises(ValueError):
        vak.prep.spectrogram_dataset.spect_helper.make_dataframe_of_spect_files(
            spect_format="npy",  # 'npy' not a valid spect format
            spect_dir=spect_dir_mat,
            annot_list=annot_list_yarden,
            annot_format="yarden",
        )


def test_make_dataframe_of_spect_files_dir_and_list_raises(
    spect_dir_mat, spect_list_mat, annot_list_yarden
):
    """test that calling ``make_dataframe_of_spect_files`` with both dir and list raises a ValueError"""
    with pytest.raises(ValueError):
        vak.prep.spectrogram_dataset.spect_helper.make_dataframe_of_spect_files(
            spect_format="mat",
            spect_dir=spect_dir_mat,
            spect_files=spect_list_mat,
            annot_list=annot_list_yarden,
            annot_format="yarden",
        )


def test_make_dataframe_of_spect_files_annot_list_without_annot_format_raises(
    spect_dir_mat, spect_list_mat, annot_list_yarden
):
    """test that calling ``make_dataframe_of_spect_files`` with a list of annotations
    but no annot_format raises a ValueError"""
    with pytest.raises(ValueError):
        vak.prep.spectrogram_dataset.spect_helper.make_dataframe_of_spect_files(
            spect_format="mat", annot_list=annot_list_yarden, annot_format=None
        )
